# This was copied from KoEDA [https://github.com/toriving/KoEDA]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

import json
import os

from kss._modules.morphemes.split_morphemes import split_morphemes
from kss._modules.josa.josa import select_josa

WORDNET_JSON_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "assets/wordnet.json"
)

with open(WORDNET_JSON_PATH, "r", encoding="utf-8") as f:
    WORDNET = json.load(f)


def get_synonyms(word, min_length=None):
    synonyms = set()
    synsets = WORDNET["lemmas"].get(word, None)

    if synsets is None:
        return []

    for syn in synsets:
        synonym = WORDNET["synsets"][syn]["lemmas"]
        if min_length is not None:
            synonym = [s for s in synonym if len(s) >= min_length]
        synonyms.update(synonym)

    if word in synonyms:
        synonyms.remove(word)

    return list(synonyms)


def correct_josa(text: str):
    morphemes = split_morphemes(text, drop_space=False)
    outputs = []
    previous = None
    for i, (word, pos) in enumerate(morphemes):
        if pos.startswith("J") and previous is not None:
            if not previous[1].startswith("S"):
                outputs.append(select_josa(previous[0], word))
            else:
                outputs.append(word)
        else:
            outputs.append(word)
        previous = (word, pos)
    return "".join(outputs)



from functools import partial
from typing import Union, List, Tuple, Dict

import xlrd

from kss._utils.logger import logger
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_num_workers

xlrd.xlsx.ensure_elementtree_imported(False, None)
xlrd.xlsx.Element_has_iter = True

from koparadigm import Paradigm, prettify

paradigm_obj = Paradigm()


def paradigm(
    text: Union[str, List[str], Tuple[str]],
    num_workers: Union[int, str] = "auto",
    verbose: bool = False,
) -> Union[Dict[str, str], List[Dict[str, str]]]:
    """
    This searches paradigms of the given text.

    Args:
        text (Union[str, List[str], Tuple[str]): single text or list of texts
        num_workers (Union[int, str]): the number of multiprocessing workers
        verbose (bool): whether to print the results or not

    Returns:
        Union[Dict[str, str], List[Dict[str, str]]]: paradigms of the given text

    Examples:
        >>> from kss import Kss
        >>> paradigm = Kss("paradigm")
        >>> text = "곱"
        >>> output = paradigm(text)
        >>> print(output)
        {'Action Verb': [('거나', '곱거나'), ('거늘', '곱거늘'), ('거니', '곱거니'), ('거니와', '곱거니와'), ('거드면', '곱거드면'), ('거든', '곱거든'), ('건', '곱건'), ('건마는', '곱건마는'), ('것다', '곱것다'), ('게', '곱게'), ('겠', '곱겠'), ('고', '곱고'), ('고도', '곱고도'), ('고말고', '곱고말고'), ('관데', '곱관데'), ('기', '곱기'), ('기로', '곱기로'), ('기로니', '곱기로니'), ('기로서', '곱기로서'), ('기로서니', '곱기로서니'), ('기로선들', '곱기로선들'), ('기에', '곱기에'), ('네', '곱네'), ('니', '곱니'), ('다', '곱다'), ('다가', '곱다가'), ('다가는', '곱다가는'), ('다간', '곱다간'), ('다니', '곱다니'), ('다마다', '곱다마다'), ('더', '곱더'), ('더구나', '곱더구나'), ('더구려', '곱더구려'), ('더군', '곱더군'), ('더냐', '곱더냐'), ('더뇨', '곱더뇨'), ('더니', '곱더니'), ('더니라', '곱더니라'), ('더니마는', '곱더니마는'), ('더니만', '곱더니만'), ('더니이까', '곱더니이까'), ('더니이다', '곱더니이다'), ('더라', '곱더라'), ('더라나', '곱더라나'), ('더라니', '곱더라니'), ('더라니까', '곱더라니까'), ('더라도', '곱더라도'), ('더라며', '곱더라며'), ('더라면', '곱더라면'), ('더라면서', '곱더라면서'), ('더라손', '곱더라손'), ('더라지', '곱더라지'), ('더람', '곱더람'), ('더이까', '곱더이까'), ('더이다', '곱더이다'), ('던', '곱던'), ('던가', '곱던가'), ('던감', '곱던감'), ('던걸', '곱던걸'), ('던고', '곱던고'), ('던데', '곱던데'), ('던바', '곱던바'), ('던지', '곱던지'), ('데', '곱데'), ('데요', '곱데요'), ('도다', '곱도다'), ('되', '곱되'), ('든', '곱든'), ('든가', '곱든가'), ('든지', '곱든지'), ('듯', '곱듯'), ('듯이', '곱듯이'), ('소이까', '곱소이까'), ('죠', '곱죠'), ('지마는', '곱지마는'), ('지만', '곱지만'), ('지요', '곱지요'), ('게', '곱게'), ('고는', '곱고는'), ('누', '곱누'), ('디', '곱디'), ('소', '곱소'), ('지', '곱지'), ('아', '곱아'), ('아도', '곱아도'), ('아라', '곱아라'), ('아서', '곱아서'), ('아야', '곱아야'), ('아야만', '곱아야만'), ('아야지', '곱아야지'), ('아요', '곱아요'), ('아지이다', '곱아지이다'), ('았', '곱았'), ('았었', '곱았었'), ('았자', '곱았자'), ('으나', '곱으나'), ('으나마', '곱으나마'), ('으니', '곱으니'), ('으니까', '곱으니까'), ('으니까는', '곱으니까는'), ('으니만치', '곱으니만치'), ('으니만큼', '곱으니만큼'), ('으려니와', '곱으려니와'), ('으려면', '곱으려면'), ('으련마는', '곱으련마는'), ('으렷다', '곱으렷다'), ('으리', '곱으리'), ('으리까', '곱으리까'), ('으리니', '곱으리니'), ('으리니라', '곱으리니라'), ('으리다', '곱으리다'), ('으리라', '곱으리라'), ('으리로다', '곱으리로다'), ('으리만치', '곱으리만치'), ('으리만큼', '곱으리만큼'), ('으리오', '곱으리오'), ('으매', '곱으매'), ('으며', '곱으며'), ('으면', '곱으면'), ('으면서', '곱으면서'), ('으므로', '곱으므로'), ('으사', '곱으사'), ('으세요', '곱으세요'), ('으셔요', '곱으셔요'), ('으소서', '곱으소서'), ('으시', '곱으시'), ('으시어요', '곱으시어요'), ('으오', '곱으오'), ('으오니까', '곱으오니까'), ('으오리까', '곱으오리까'), ('으오리다', '곱으오리다'), ('으오리이까', '곱으오리이까'), ('으오리이다', '곱으오리이다'), ('으오이다', '곱으오이다'), ('으옵', '곱으옵'), ('으옵나이까', '곱으옵나이까'), ('으옵나이다', '곱으옵나이다'), ('으옵니까', '곱으옵니까'), ('으옵니다', '곱으옵니다'), ('으옵디까', '곱으옵디까'), ('으옵디다', '곱으옵디다'), ('으옵소서', '곱으옵소서'), ('으옵시', '곱으옵시'), ('으와', '곱으와'), ('으우', '곱으우'), ('은', '곱은'), ('은걸', '곱은걸'), ('은들', '곱은들'), ('은즉', '곱은즉'), ('은즉슨', '곱은즉슨'), ('을', '곱을'), ('을걸', '곱을걸'), ('을까', '곱을까'), ('을는지', '곱을는지'), ('을라', '곱을라'), ('을라고', '곱을라고'), ('을러니', '곱을러니'), ('을러라', '곱을러라'), ('을런가', '곱을런가'), ('을런고', '곱을런고'), ('을레', '곱을레'), ('을레라', '곱을레라'), ('을망정', '곱을망정'), ('을밖에', '곱을밖에'), ('을뿐더러', '곱을뿐더러'), ('을새', '곱을새'), ('을세', '곱을세'), ('을세라', '곱을세라'), ('을세말이지', '곱을세말이지'), ('을수록', '곱을수록'), ('을시', '곱을시'), ('을쏘냐', '곱을쏘냐'), ('을쏜가', '곱을쏜가'), ('을지', '곱을지'), ('을지나', '곱을지나'), ('을지니', '곱을지니'), ('을지니라', '곱을지니라'), ('을지라', '곱을지라'), ('을지라도', '곱을지라도'), ('을지로다', '곱을지로다'), ('을지며', '곱을지며'), ('을지어다', '곱을지어다'), ('을지언정', '곱을지언정'), ('을진대', '곱을진대'), ('을진대는', '곱을진대는'), ('을진저', '곱을진저'), ('음', '곱음'), ('음에도', '곱음에도'), ('음에랴', '곱음에랴'), ('읍시오', '곱읍시오'), ('습네', '곱습네'), ('습늰다', '곱습늰다'), ('습니까', '곱습니까'), ('습니다', '곱습니다'), ('습디까', '곱습디까'), ('습디다', '곱습디다'), ('습딘다', '곱습딘다'), ('습죠', '곱습죠'), ('습지요', '곱습지요'), ('거라', '곱거라'), ('건대', '곱건대'), ('게', '곱게'), ('게끔', '곱게끔'), ('게나', '곱게나'), ('고는', '곱고는'), ('고서', '곱고서'), ('고야', '곱고야'), ('곤', '곱곤'), ('구려', '곱구려'), ('구먼', '곱구먼'), ('나', '곱나'), ('나니', '곱나니'), ('남', '곱남'), ('노니', '곱노니'), ('노라', '곱노라'), ('노라고', '곱노라고'), ('누나', '곱누나'), ('누먼', '곱누먼'), ('느라', '곱느라'), ('느라고', '곱느라고'), ('는구나', '곱는구나'), ('는구려', '곱는구려'), ('는군', '곱는군'), ('도록', '곱도록'), ('세나', '곱세나'), ('자', '곱자'), ('자고', '곱자고'), ('자꾸나', '곱자꾸나'), ('자느니', '곱자느니'), ('자니까', '곱자니까'), ('자마자', '곱자마자'), ('자며', '곱자며'), ('자면', '곱자면'), ('자면서', '곱자면서'), ('자손', '곱자손'), ('아다', '곱아다'), ('아다가', '곱아다가'), ('고자', '곱고자'), ('나이까', '곱나이까'), ('나이다', '곱나이다'), ('노라니', '곱노라니'), ('노라니까', '곱노라니까'), ('노라면', '곱노라면'), ('느냐', '곱느냐'), ('느냐고', '곱느냐고'), ('느뇨', '곱느뇨'), ('느니', '곱느니'), ('느니라', '곱느니라'), ('느니만', '곱느니만'), ('느니만치', '곱느니만치'), ('느니만큼', '곱느니만큼'), ('는', '곱는'), ('는가', '곱는가'), ('는감', '곱는감'), ('는걸', '곱는걸'), ('는고', '곱는고'), ('는데', '곱는데'), ('는뎁쇼', '곱는뎁쇼'), ('는바', '곱는바'), ('는지', '곱는지'), ('는지고', '곱는지고'), ('는지라', '곱는지라'), ('으라', '곱으라'), ('으라고', '곱으라고'), ('으라나', '곱으라나'), ('으라느니', '곱으라느니'), ('으라니까', '곱으라니까'), ('으라며', '곱으라며'), ('으라면서', '곱으라면서'), ('으라손', '곱으라손'), ('으락', '곱으락'), ('으란', '곱으란'), ('으람', '곱으람'), ('으랴', '곱으랴'), ('으러', '곱으러'), ('으려', '곱으려'), ('으려거든', '곱으려거든'), ('으려고', '곱으려고'), ('으려나', '곱으려나'), ('으려니', '곱으려니'), ('으려든', '곱으려든'), ('으려무나', '곱으려무나'), ('으련', '곱으련'), ('으렴', '곱으렴'), ('으렵니까', '곱으렵니까'), ('으렵니다', '곱으렵니다'), ('으마', '곱으마'), ('으사이다', '곱으사이다'), ('으세', '곱으세'), ('으시라', '곱으시라'), ('으시압', '곱으시압'), ('으십사', '곱으십사'), ('으십시다', '곱으십시다'), ('으십시오', '곱으십시오'), ('을거나', '곱을거나'), ('을게', '곱을게'), ('을깝쇼', '곱을깝쇼'), ('을라치면', '곱을라치면'), ('을락', '곱을락'), ('을래', '곱을래'), ('을작시면', '곱을작시면'), ('음세', '곱음세'), ('읍시다', '곱읍시다'), ('읍시사', '곱읍시사'), ('는다', '곱는다'), ('는다고', '곱는다고'), ('는다나', '곱는다나'), ('는다네', '곱는다네'), ('는다느니', '곱는다느니'), ('는다니', '곱는다니'), ('는다니까', '곱는다니까'), ('는다더라', '곱는다더라'), ('는다마는', '곱는다마는'), ('는다며', '곱는다며'), ('는다면', '곱는다면'), ('는다면서', '곱는다면서'), ('는다손', '곱는다손'), ('는다오', '곱는다오'), ('는다지', '곱는다지'), ('는단다', '곱는단다'), ('는담', '곱는담'), ('는답니까', '곱는답니까'), ('는답니다', '곱는답니다'), ('는답시고', '곱는답시고'), ('는대', '곱는대'), ('는대요', '곱는대요')], 'Descriptive Verb': [('거나', '곱거나'), ('거늘', '곱거늘'), ('거니', '곱거니'), ('거니와', '곱거니와'), ('거드면', '곱거드면'), ('거든', '곱거든'), ('건', '곱건'), ('건마는', '곱건마는'), ('것다', '곱것다'), ('게', '곱게'), ('겠', '곱겠'), ('고', '곱고'), ('고도', '곱고도'), ('고말고', '곱고말고'), ('관데', '곱관데'), ('기', '곱기'), ('기로', '곱기로'), ('기로니', '곱기로니'), ('기로서', '곱기로서'), ('기로서니', '곱기로서니'), ('기로선들', '곱기로선들'), ('기에', '곱기에'), ('네', '곱네'), ('니', '곱니'), ('다', '곱다'), ('다가', '곱다가'), ('다가는', '곱다가는'), ('다간', '곱다간'), ('다니', '곱다니'), ('다마다', '곱다마다'), ('더', '곱더'), ('더구나', '곱더구나'), ('더구려', '곱더구려'), ('더군', '곱더군'), ('더냐', '곱더냐'), ('더뇨', '곱더뇨'), ('더니', '곱더니'), ('더니라', '곱더니라'), ('더니마는', '곱더니마는'), ('더니만', '곱더니만'), ('더니이까', '곱더니이까'), ('더니이다', '곱더니이다'), ('더라', '곱더라'), ('더라나', '곱더라나'), ('더라니', '곱더라니'), ('더라니까', '곱더라니까'), ('더라도', '곱더라도'), ('더라며', '곱더라며'), ('더라면', '곱더라면'), ('더라면서', '곱더라면서'), ('더라손', '곱더라손'), ('더라지', '곱더라지'), ('더람', '곱더람'), ('더이까', '곱더이까'), ('더이다', '곱더이다'), ('던', '곱던'), ('던가', '곱던가'), ('던감', '곱던감'), ('던걸', '곱던걸'), ('던고', '곱던고'), ('던데', '곱던데'), ('던바', '곱던바'), ('던지', '곱던지'), ('데', '곱데'), ('데요', '곱데요'), ('도다', '곱도다'), ('되', '곱되'), ('든', '곱든'), ('든가', '곱든가'), ('든지', '곱든지'), ('듯', '곱듯'), ('듯이', '곱듯이'), ('소이까', '곱소이까'), ('죠', '곱죠'), ('지마는', '곱지마는'), ('지만', '곱지만'), ('지요', '곱지요'), ('게', '곱게'), ('고는', '곱고는'), ('누', '곱누'), ('디', '곱디'), ('소', '곱소'), ('지', '곱지'), ('아', '고와'), ('아도', '고와도'), ('아라', '고와라'), ('아서', '고와서'), ('아야', '고와야'), ('아야만', '고와야만'), ('아야지', '고와야지'), ('아요', '고와요'), ('아지이다', '고와지이다'), ('았', '고왔'), ('았었', '고왔었'), ('았자', '고왔자'), ('으나', '고우나'), ('으나마', '고우나마'), ('으니', '고우니'), ('으니까', '고우니까'), ('으니까는', '고우니까는'), ('으니만치', '고우니만치'), ('으니만큼', '고우니만큼'), ('으려니와', '고우려니와'), ('으려면', '고우려면'), ('으련마는', '고우련마는'), ('으렷다', '고우렷다'), ('으리', '고우리'), ('으리까', '고우리까'), ('으리니', '고우리니'), ('으리니라', '고우리니라'), ('으리다', '고우리다'), ('으리라', '고우리라'), ('으리로다', '고우리로다'), ('으리만치', '고우리만치'), ('으리만큼', '고우리만큼'), ('으리오', '고우리오'), ('으매', '고우매'), ('으며', '고우며'), ('으면', '고우면'), ('으면서', '고우면서'), ('으므로', '고우므로'), ('으사', '고우사'), ('으세요', '고우세요'), ('으셔요', '고우셔요'), ('으소서', '고우소서'), ('으시', '고우시'), ('으시어요', '고우시어요'), ('으오', '고우오'), ('으오니까', '고우오니까'), ('으오리까', '고우오리까'), ('으오리다', '고우오리다'), ('으오리이까', '고우오리이까'), ('으오리이다', '고우오리이다'), ('으오이다', '고우오이다'), ('으옵', '고우옵'), ('으옵나이까', '고우옵나이까'), ('으옵나이다', '고우옵나이다'), ('으옵니까', '고우옵니까'), ('으옵니다', '고우옵니다'), ('으옵디까', '고우옵디까'), ('으옵디다', '고우옵디다'), ('으옵소서', '고우옵소서'), ('으옵시', '고우옵시'), ('으와', '고우와'), ('으우', '고우우'), ('은', '고운'), ('은걸', '고운걸'), ('은들', '고운들'), ('은즉', '고운즉'), ('은즉슨', '고운즉슨'), ('을', '고울'), ('을걸', '고울걸'), ('을까', '고울까'), ('을는지', '고울는지'), ('을라', '고울라'), ('을라고', '고울라고'), ('을러니', '고울러니'), ('을러라', '고울러라'), ('을런가', '고울런가'), ('을런고', '고울런고'), ('을레', '고울레'), ('을레라', '고울레라'), ('을망정', '고울망정'), ('을밖에', '고울밖에'), ('을뿐더러', '고울뿐더러'), ('을새', '고울새'), ('을세', '고울세'), ('을세라', '고울세라'), ('을세말이지', '고울세말이지'), ('을수록', '고울수록'), ('을시', '고울시'), ('을쏘냐', '고울쏘냐'), ('을쏜가', '고울쏜가'), ('을지', '고울지'), ('을지나', '고울지나'), ('을지니', '고울지니'), ('을지니라', '고울지니라'), ('을지라', '고울지라'), ('을지라도', '고울지라도'), ('을지로다', '고울지로다'), ('을지며', '고울지며'), ('을지어다', '고울지어다'), ('을지언정', '고울지언정'), ('을진대', '고울진대'), ('을진대는', '고울진대는'), ('을진저', '고울진저'), ('음', '고움'), ('음에도', '고움에도'), ('음에랴', '고움에랴'), ('읍시오', '고웁시오'), ('습네', '곱습네'), ('습늰다', '곱습늰다'), ('습니까', '곱습니까'), ('습니다', '곱습니다'), ('습디까', '곱습디까'), ('습디다', '곱습디다'), ('습딘다', '곱습딘다'), ('습죠', '곱습죠'), ('습지요', '곱습지요'), ('구나', '곱구나'), ('구려', '곱구려'), ('군', '곱군'), ('다고', '곱다고'), ('다나', '곱다나'), ('다네', '곱다네'), ('다느니', '곱다느니'), ('다니', '곱다니'), ('다니까', '곱다니까'), ('다더라', '곱다더라'), ('다마는', '곱다마는'), ('다며', '곱다며'), ('다면', '곱다면'), ('다면서', '곱다면서'), ('다손', '곱다손'), ('다지', '곱다지'), ('단다', '곱단다'), ('담', '곱담'), ('답니까', '곱답니까'), ('답니다', '곱답니다'), ('답시고', '곱답시고'), ('대', '곱대'), ('대요', '곱대요'), ('디', '곱디'), ('우', '곱우'), ('으냐', '고우냐'), ('으냐고', '고우냐고'), ('으뇨', '고우뇨'), ('으니', '고우니'), ('으니라', '고우니라'), ('으이', '고우이'), ('은가', '고운가'), ('은감', '고운감'), ('은고', '고운고'), ('은데', '고운데'), ('은뎁쇼', '고운뎁쇼'), ('은지', '고운지'), ('을데라니', '고울데라니'), ('을시고', '고울시고')]}

    References:
        This was copied from [KoParadigm](https://github.com/Kyubyong/KoParadigm) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return [] if isinstance(text, list) else {}

    num_workers = _check_num_workers(text, num_workers)

    if num_workers is not False and verbose:
        verbose = False
        logger.warn(
            "Verbose mode is not supported for multiprocessing. "
            "It will be turned off automatically."
        )

    return _run_job(
        func=partial(_paradigm, verbose=verbose),
        inputs=text,
        num_workers=num_workers,
    )


def _paradigm(text: str, verbose: bool = False):
    outputs = paradigm_obj.conjugate(text)

    if not outputs:
        return {}

    if verbose:
        prettify(outputs)

    outputs_dict = {}
    for output in outputs:
        outputs_dict[output[0]] = output[1]

    return outputs_dict



def test_summarize_sentences():
    from kss import Kss
    summarize_sentences = Kss("summarize_sentences")
    text = "개그맨 겸 가수 ‘개가수’ UV 유세윤이 신곡 발매 이후 많은 남편들의 응원을 받고 있다. 유세윤은 지난 3일 오후 6시 새 싱글 ‘마더 사커(Mother Soccer)(Feat. 수퍼비)’를 발매했다. ‘마더 사커’는 아내에 대한 서운한 마음을 위트 있고 강한 어조로 디스 하는 남편 유세윤의 마음을 담은 곡이다. 발매 후 소셜 미디어 상에서 화제를 모으고 있는 가운데, 가수 하동균은 “유세유니 괜찮겠어”라는 반응을 보이기도 했다. 누리꾼들은 ‘두 분의 원만한 합의가 있기를 바랍니다’, ‘집에는 들어갈 수 있겠나’ 등 유세윤의 귀가를 걱정하는 모습을 보였다. 유세윤은 점입가경으로 ‘마더 사커’ 챌린지를 시작, 자신의 SNS를 통해 “부부 싸움이 좀 커졌네요”라며 배우 송진우와 함께 촬영한 영상을 게재했다. 해당 영상에서는 양말을 신고 침대에 들어간 뒤 환호를 지르거나 화장실 불을 끄지 않고 도망가는 등 아내의 잔소리 유발 포인트를 살려 재치 있는 영상을 완성했다. 유세윤은 ‘마더 사커’를 통해 남편들의 마음을 대변해 주고 있는 한편 아내의 반응은 어떨지 궁금증을 모은다."
    output = summarize_sentences(text)
    assert output == ['개그맨 겸 가수 ‘개가수’ UV 유세윤이 신곡 발매 이후 많은 남편들의 응원을 받고 있다.',
                      '‘마더 사커’는 아내에 대한 서운한 마음을 위트 있고 강한 어조로 디스 하는 남편 유세윤의 마음을 담은 곡이다.',
                      '유세윤은 ‘마더 사커’를 통해 남편들의 마음을 대변해 주고 있는 한편 아내의 반응은 어떨지 궁금증을 모은다.']



# Copyright (C) 2021 Hyunwoong Ko <kevin.brain@kakaobrain.com> and Sang-Kil Park <skpark1224@hyundai.com>
# All rights reserved.

from kss._modules.augmentation.augment import augment
from kss._modules.collocation.collocate import collocate
from kss._modules.g2p.g2p import g2p
from kss._modules.hangulization.hangulization import hangulize
from kss._modules.hanja.hanja import split_hanja, is_hanja, hanja2hangul
from kss._modules.jamo._jamo import h2j, h2hcj, j2h, j2hcj, hcj2h, hcj2j, is_jamo, is_jamo_modern, is_hcj, \
    is_hcj_modern, is_hangul_char
from kss._modules.josa.josa import select_josa, combine_josa
from kss._modules.keywords.extract_keywords import extract_keywords
from kss._modules.morphemes.split_morphemes import split_morphemes
from kss._modules.paradigm.paradigm import paradigm
from kss._modules.preprocessing.anonymize import anonymize
from kss._modules.preprocessing.clean_news import clean_news
from kss._modules.preprocessing.completed_form import is_completed_form, get_all_completed_form_hangul_chars, \
    get_all_incompleted_form_hangul_chars
from kss._modules.preprocessing.filter_out import filter_out
from kss._modules.preprocessing.half2full import half2full
from kss._modules.preprocessing.normalize import normalize
from kss._modules.preprocessing.preprocess import preprocess
from kss._modules.preprocessing.reduce_repeats import reduce_char_repeats, reduce_emoticon_repeats
from kss._modules.preprocessing.remove_invisible_chars import remove_invisible_chars
from kss._modules.qwerty.qwerty import qwerty
from kss._modules.romanization.romanize import romanize
from kss._modules.safety.check_safety import is_unsafe
from kss._modules.sentences.split_sentences import split_sentences
from kss._modules.spacing.correct_spacing import correct_spacing
from kss._modules.summarization.summarize_sentences import summarize_sentences

supported_modules = {
    "augment": augment,
    "collocate": collocate,
    "g2p": g2p,
    "hangulize": hangulize,
    "split_hanja": split_hanja,
    "is_hanja": is_hanja,
    "hanja2hangul": hanja2hangul,
    "h2j": h2j,
    "h2hcj": h2hcj,
    "j2h": j2h,
    "j2hcj": j2hcj,
    "hcj2h": hcj2h,
    "hcj2j": hcj2j,
    "is_jamo": is_jamo,
    "is_jamo_modern": is_jamo_modern,
    "is_hcj": is_hcj,
    "is_hcj_modern": is_hcj_modern,
    "is_hangul_char": is_hangul_char,
    "select_josa": select_josa,
    "combine_josa": combine_josa,
    "extract_keywords": extract_keywords,
    "split_morphemes": split_morphemes,
    "paradigm": paradigm,
    "anonymize": anonymize,
    "clean_news": clean_news,
    "is_completed_form": is_completed_form,
    "get_all_completed_form_hangul_chars": get_all_completed_form_hangul_chars,
    "get_all_incompleted_form_hangul_chars": get_all_incompleted_form_hangul_chars,
    "filter_out": filter_out,
    "half2full": half2full,
    "reduce_char_repeats": reduce_char_repeats,
    "reduce_emoticon_repeats": reduce_emoticon_repeats,
    "remove_invisible_chars": remove_invisible_chars,
    "normalize": normalize,
    "preprocess": preprocess,
    "qwerty": qwerty,
    "romanize": romanize,
    "is_unsafe": is_unsafe,
    "split_sentences": split_sentences,
    "correct_spacing": correct_spacing,
    "summarize_sentences": summarize_sentences,
}

alias = {
    "aug": "augment",
    "augmentation": "augment",
    "collocation": "collocate",
    "hangulization": "hangulize",
    "hangulisation": "hangulize",
    "hangulise": "hangulize",
    "hanja": "hanja2hangul",
    "hangul2jamo": "h2j",
    "hangul2hcj": "h2hcj",
    "jamo2hangul": "j2h",
    "jamo2hcj": "j2hcj",
    "hcj2hangul": "hcj2h",
    "hcj2jamo": "hcj2j",
    "josa": "select_josa",
    "keyword": "extract_keywords",
    "keywords": "extract_keywords",
    "morpheme": "split_morphemes",
    "morphemes": "split_morphemes",
    "annonymization": "anonymize",
    "news_cleaning": "clean_news",
    "news": "clean_news",
    "completed_form": "is_completed_form",
    "completed": "is_completed_form",
    "filter": "filter_out",
    "reduce_repeats": "reduce_char_repeats",
    "reduce_char": "reduce_char_repeats",
    "reduce_chars": "reduce_char_repeats",
    "reduce_emoticon": "reduce_emoticon_repeats",
    "reduce_emoticons": "reduce_emoticon_repeats",
    "reduce_emo": "reduce_emoticon_repeats",
    "remove_invisible": "remove_invisible_chars",
    "invisible_chars": "remove_invisible_chars",
    "invisible": "remove_invisible_chars",
    "normalization": "normalize",
    "normalisation": "normalize",
    "normalise": "normalize",
    "preprocessing": "preprocess",
    "prep": "preprocess",
    "romanization": "romanize",
    "romanisation": "romanize",
    "romanise": "romanize",
    "safety": "is_unsafe",
    "check_safety": "is_unsafe",
    "sentence": "split_sentences",
    "sentences": "split_sentences",
    "sent_split": "split_sentences",
    "sent_splits": "split_sentences",
    "sents_split": "split_sentences",
    "split_sent": "split_sentences",
    "split_sents": "split_sentences",
    "spacing": "correct_spacing",
    "space": "correct_spacing",
    "spaces": "correct_spacing",
    "summarization": "summarize_sentences",
    "summarize": "summarize_sentences",
    "summ": "summarize_sentences",
    "morph": "split_morphemes",
    "morphs": "split_morphemes",
    "tokenize": "split_morphemes",
    "tokenization": "split_morphemes",
    "split_morph": "split_morphemes",
    "split_morphs": "split_morphemes",
    "morph_split": "split_morphemes",
    "morph_splits": "split_morphemes",
    "morphs_split": "split_morphemes",
}


class Kss(object):
    def __init__(self, module: str):
        self.module = self._check_module(module, supported_modules, alias)

    def __call__(self, *args, **kwargs):
        return self.module(*args, **kwargs)

    def help(self):
        print(self.module.__doc__.strip())

    @staticmethod
    def available():
        return list(supported_modules.keys())

    @staticmethod
    def alias():
        return alias

    def _check_module(self, module: str, supported_modules, alias):
        from kss._utils.sanity_checks import _check_type

        module = module.lower()
        module = _check_type(module, "module", str)

        if module in supported_modules:
            return supported_modules[module]

        elif module in alias:
            return supported_modules[alias[module]]

        else:
            error_message = f"'{module}' is not supported module." \
                            f"\nSupported modules are the following:" \
                            f"\n{list(supported_modules.keys())}"
            closest_module = self._find_closest_module(module)
            if closest_module:
                if closest_module in alias:
                    closest_module = alias[closest_module]
                error_message += f"\n\nDid you mean '{closest_module}'?"
            raise ValueError(error_message)

    @staticmethod
    def _find_closest_module(module, min_distance=0.5):
        import distance
        current_min_distance = 99
        closest_module = None
        for supported_module in list(supported_modules.keys()) + list(alias.keys()):
            dist = distance.nlevenshtein(module, supported_module)
            if dist < current_min_distance and dist <= min_distance:
                current_min_distance = dist
                closest_module = supported_module

        if closest_module is None:
            return None
        else:
            return closest_module


__ALL__ = list(supported_modules.keys()) + ["Kss"]
__version__ = "6.0.3"



from functools import partial, lru_cache
from typing import Union, List, Tuple

from kss._modules.qwerty.utils import (
    HANGUL,
    HANGUL_FIRST,
    HANGUL_LAST,
    ENGLISH,
    LEADS,
    VOWELS,
    TAILS,
    CONSONANT_FIRST,
    VOWEL_LAST,
    CONNECTABLE_CONSONANTS,
    CONNECTABLE_VOWELS,
    ENGLISH_INDEX,
    separate_hangul,
    generate_hangul,
    index_of,
    is_vowel,
)
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_type, _check_num_workers


def qwerty(
    text: Union[str, List[str], Tuple[str]],
    src: str,
    tgt: str,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This converts text from one language to another using QWERTY keyboard layout.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        src (str): source language
        tgt (str): target language
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: converted text or list of converted texts

    Examples:
        >>> from kss import Kss
        >>> qwerty = Kss("qwerty")
        >>> text = "dkssudgktpdy"
        >>> qwerty(text, src="en", tgt="ko")
        '안녕하세요'
        >>> text = "안녕하세요"
        >>> qwerty(text, src="ko", tgt="en")
        'dkssudgktpdy'

    References:
        This was copied from [inko.py](https://github.com/738/inko.py) and modified by Kss
    """
    text, finish = _check_text(text)

    if finish:
        return text

    src = _check_type(src, "src", str)
    tgt = _check_type(tgt, "tgt", str)
    num_workers = _check_num_workers(text, num_workers)

    if src not in ["en", "ko"] or tgt not in ["en", "ko"]:
        raise ValueError("`src` and `tgt` must be one of 'en', 'ko'")

    return _run_job(
        func=partial(_qwerty, src=src, tgt=tgt),
        inputs=text,
        num_workers=num_workers,
    )


@lru_cache(maxsize=500)
def _qwerty(text, src, tgt):
    if src == tgt:
        return text

    if src == "en" and tgt == "ko":
        return _qwerty_en2ko(text)
    elif src == "ko" and tgt == "en":
        return _qwerty_ko2en(text)
    else:
        return text


def _qwerty_en2ko(text, allow_double_consonant=True):
    state_length = [0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 5]
    transitions = [
        [1, 1, 2, 2],  # 0, EMPTY
        [3, 1, 4, 4],  # 1, 자
        [1, 1, 5, 2],  # 2, 모
        [3, 1, 4, -1],  # 3, 자자
        [6, 1, 7, 2],  # 4, 자모
        [1, 1, 2, 2],  # 5, 모모
        [9, 1, 4, 4],  # 6, 자모자
        [9, 1, 2, 2],  # 7, 자모모
        [1, 1, 4, 4],  # 8, 자모자자
        [10, 1, 4, 4],  # 9, 자모모자
        [1, 1, 4, 4],  # 10, 자모모자자
    ]

    def last(_list):
        return _list[len(_list) - 1]

    def combine(arr):
        group = []
        for i in range(len(arr)):
            h = HANGUL[arr[i]]
            if i == 0 or is_vowel(last(group)[0]) != is_vowel(h):
                group.append([])
            last(group).append(h)

        def connect(e):
            w = "".join(e)
            if w in CONNECTABLE_CONSONANTS:
                return CONNECTABLE_CONSONANTS[w]
            elif w in CONNECTABLE_VOWELS:
                return CONNECTABLE_VOWELS[w]
            else:
                return w

        group = [connect(e) for e in group]

        if len(group) == 1:
            return group[0]

        char_set = [LEADS, VOWELS, TAILS]

        code = []
        try:
            code = [index_of(w, char_set[i]) for i, w in enumerate(group)]
        except IndexError:
            pass

        if len(code) < 3:
            code.append(-1)

        return generate_hangul(code)

    def finalize():
        length = len(text)
        last = -1
        result = []
        state = 0
        temp_dict = {
            "tmp": []
        }

        def flush():
            if len(temp_dict["tmp"]) > 0:
                result.append(combine(temp_dict["tmp"]))
            temp_dict["tmp"] = []

        for i in range(length):
            char = text[i]
            if char not in ENGLISH_INDEX:
                state = 0
                flush()
                result.append(char)
            else:
                curr = ENGLISH_INDEX[char]

                def transition():
                    c = (HANGUL[last] if last > -1 else "") + HANGUL[curr]
                    last_is_vowel = is_vowel(HANGUL[last])
                    curr_is_vowel = is_vowel(HANGUL[curr])
                    if not curr_is_vowel:
                        if last_is_vowel:
                            return 0 if index_of("ㄸㅃㅉ", HANGUL[curr]) == -1 else 1
                        if state == 1 and not allow_double_consonant:
                            return 1
                        return 0 if c in CONNECTABLE_CONSONANTS else 1
                    elif last_is_vowel:
                        return 2 if c in CONNECTABLE_VOWELS else 3

                    return 2

                _transition = transition()
                next_state = transitions[state][_transition]
                temp_dict["tmp"].append(curr)
                diff = len(temp_dict["tmp"]) - state_length[next_state]

                if diff > 0:
                    result.append(combine(temp_dict["tmp"][0:diff]))
                    temp_dict["tmp"] = temp_dict["tmp"][diff:]
                state = next_state
                last = curr

        flush()
        return "".join(result)

    return finalize()


def _qwerty_ko2en(text):
    result = ""
    if text == "" or text is None:
        return result
    _seperated = [-1, -1, -1, -1, -1]

    for i in range(len(text)):
        _hangul_text = text[i]
        _code = ord(_hangul_text)

        if (HANGUL_FIRST <= _code <= HANGUL_LAST) or (CONSONANT_FIRST <= _code <= VOWEL_LAST):
            # 가 ~ 힣 사이에 있는 한글일 때
            _seperated = separate_hangul(_hangul_text)
        else:
            # 한글이 아닐 때
            result += _hangul_text
            _seperated = [-1, -1, -1, -1, -1]

        for j in range(len(_seperated)):
            if _seperated[j] != -1:
                result += ENGLISH[_seperated[j]]

    return result



# This code was copied from g2pk [https://github.com/kyubyong/g2pK]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko]

import os
import re

import cmudict

import kss
from kss._modules.jamo._jamo import h2j, j2h


def adjust(arpabets):
    """Modify arpabets so that it fits our processes"""
    string = " " + " ".join(arpabets) + " $"
    string = re.sub(r"\d", "", string)
    string = string.replace(" T S ", " TS ")
    string = string.replace(" D Z ", " DZ ")
    string = string.replace(" AW ER ", " AWER ")
    string = string.replace(" IH R $", " IH ER ")
    string = string.replace(" EH R $", " EH ER ")
    string = string.replace(" $", "")
    return string.strip("$ ").split()


def to_choseong(arpabet):
    """Arpabet to choseong or onset"""
    d = \
        {'B': 'ᄇ',
         'CH': 'ᄎ',
         'D': 'ᄃ',
         'DH': 'ᄃ',
         'DZ': 'ᄌ',
         'F': 'ᄑ',
         'G': 'ᄀ',
         'HH': 'ᄒ',
         'JH': 'ᄌ',
         'K': 'ᄏ',
         'L': 'ᄅ',
         'M': 'ᄆ',
         'N': 'ᄂ',
         'NG': 'ᄋ',
         'P': 'ᄑ',
         'R': 'ᄅ',
         'S': 'ᄉ',
         'SH': 'ᄉ',
         'T': 'ᄐ',
         'TH': 'ᄉ',
         'TS': 'ᄎ',
         'V': 'ᄇ',
         'W': 'W',
         'Y': 'Y',
         'Z': 'ᄌ',
         'ZH': 'ᄌ'}

    return d.get(arpabet, arpabet)


def to_jungseong(arpabet):
    """Arpabet to jungseong or vowel"""
    d = \
        {'AA': 'ᅡ',
         'AE': 'ᅢ',
         'AH': 'ᅥ',
         'AO': 'ᅩ',
         'AW': 'ᅡ우',
         'AWER': "ᅡ워",
         'AY': 'ᅡ이',
         'EH': 'ᅦ',
         'ER': 'ᅥ',
         'EY': 'ᅦ이',
         'IH': 'ᅵ',
         'IY': 'ᅵ',
         'OW': 'ᅩ',
         'OY': 'ᅩ이',
         'UH': 'ᅮ',
         'UW': 'ᅮ'}
    return d.get(arpabet, arpabet)


def to_jongseong(arpabet):
    """Arpabet to jongseong or coda"""
    d = \
        {'B': 'ᆸ',
         'CH': 'ᆾ',
         'D': 'ᆮ',
         'DH': 'ᆮ',
         'F': 'ᇁ',
         'G': 'ᆨ',
         'HH': 'ᇂ',
         'JH': 'ᆽ',
         'K': 'ᆨ',
         'L': 'ᆯ',
         'M': 'ᆷ',
         'N': 'ᆫ',
         'NG': 'ᆼ',
         'P': 'ᆸ',
         'R': 'ᆯ',
         'S': 'ᆺ',
         'SH': 'ᆺ',
         'T': 'ᆺ',
         'TH': 'ᆺ',
         'V': 'ᆸ',
         'W': 'ᆼ',
         'Y': 'ᆼ',
         'Z': 'ᆽ',
         'ZH': 'ᆽ'}

    return d.get(arpabet, arpabet)


def reconstruct(string):
    """Some postprocessing rules"""
    pairs = [("그W", "ᄀW"),
             ("흐W", "ᄒW"),
             ("크W", "ᄏW"),
             ("ᄂYᅥ", "니어"),
             ("ᄃYᅥ", "디어"),
             ("ᄅYᅥ", "리어"),
             ("Yᅵ", "ᅵ"),
             ("Yᅡ", "ᅣ"),
             ("Yᅢ", "ᅤ"),
             ("Yᅥ", "ᅧ"),
             ("Yᅦ", "ᅨ"),
             ("Yᅩ", "ᅭ"),
             ("Yᅮ", "ᅲ"),
             ("Wᅡ", "ᅪ"),
             ("Wᅢ", "ᅫ"),
             ("Wᅥ", "ᅯ"),
             ("Wᅩ", "ᅯ"),
             ("Wᅮ", "ᅮ"),
             ("Wᅦ", "ᅰ"),
             ("Wᅵ", "ᅱ"),
             ("ᅳᅵ", "ᅴ"),
             ("Y", "ᅵ"),
             ("W", "ᅮ")
             ]
    for str1, str2 in pairs:
        string = string.replace(str1, str2)
    return string


def parse_table():
    """Parse the main rule table"""
    file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets", "table.csv")
    lines = open(file_path, 'r', encoding='utf8').read().splitlines()
    onsets = lines[0].split(",")
    table = []
    for line in lines[1:]:
        cols = line.split(",")
        coda = cols[0]
        for i, onset in enumerate(onsets):
            cell = cols[i]
            if len(cell) == 0: continue
            if i == 0:
                continue
            else:
                str1 = f"{coda}{onset}"
                if "(" in cell:
                    str2 = cell.split("(")[0]
                    rule_ids = cell.split("(")[1][:-1].split("/")
                else:
                    str2 = cell
                    rule_ids = []

                table.append((str1, str2, rule_ids))
    return table


def annotate(string, backend):
    """attach pos tags to the given string using Mecab"""
    tokens = kss.split_morphemes(string, backend=backend)

    if string.replace(" ", "") != "".join(token for token, _ in tokens):
        return string
    blanks = [i for i, char in enumerate(string) if char == " "]

    tag_seq = []
    for token, tag in tokens:
        tag = tag.split("+")[-1]
        if tag == "NNBC":  # bound noun
            tag = "B"
        else:
            tag = tag[0]
        tag_seq.append("_" * (len(token) - 1) + tag)
    tag_seq = "".join(tag_seq)

    for i in blanks:
        tag_seq = tag_seq[:i] + " " + tag_seq[i:]

    annotated = ""
    for char, tag in zip(string, tag_seq):
        annotated += char
        if char == "의" and tag == "J":
            annotated += "/J"
        elif tag == "E":
            if h2j(char)[-1] in "ᆯ":
                annotated += "/E"
        elif tag == "V":
            if h2j(char)[-1] in "ᆫᆬᆷᆱᆰᆲᆴ":
                annotated += "/P"
        elif tag == "B":  # bound noun
            annotated += "/B"

    return annotated


def group(inp):
    """For group_vowels=True
    Contemporarily, Korean speakers don't distinguish some vowels.
    """
    inp = inp.replace("ᅢ", "ᅦ")
    inp = inp.replace("ᅤ", "ᅨ")
    inp = inp.replace("ᅫ", "ᅬ")
    inp = inp.replace("ᅰ", "ᅬ")
    return inp


def _get_examples():
    """For internal use"""
    file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets", "rules.txt")
    text = open(file_path, encoding='utf8').readlines()

    examples = []
    for line in text:
        if line.startswith("->"):
            examples.extend(re.findall(r"([ㄱ-힣][ ㄱ-힣]*)\[([ㄱ-힣][ ㄱ-힣]*)]", line))

    _examples = []
    for inp, gt in examples:
        for each in gt.split("/"):
            _examples.append((inp, each))

    return _examples


def get_rule_id2text():
    """for verbose=True"""
    file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets", "rules.txt")
    rules = open(file_path, encoding='utf8').read().strip().split("\n\n")
    rule_id2text = dict()
    for rule in rules:
        rule_id, texts = rule.splitlines()[0], rule.splitlines()[1:]
        rule_id2text[rule_id.strip()] = "\n".join(texts)
    return rule_id2text


def get_idioms():
    file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets", "idioms.txt")
    return open(file_path, 'r', encoding="utf8").readlines()


def gloss(verbose, out, inp, rule):
    """displays the process and relevant information"""
    if verbose and out != inp and out != re.sub("/[EJPB]", "", inp):
        print(j2h(inp), "->", j2h(out))
        print(f"\033[1;31m{rule}\033[0m")


def convert_idioms(
    text,
    descriptive=False,
    convert_english_to_hangul_phonemes=True,
    convert_numbers_to_hangul_phonemes=True,
    verbose=False,
):
    """Process each line in `idioms.txt`
    Each line is delimited by "===",
    and the left string is replaced by the right one.
    inp: input string.
    descriptive: not used.
    verbose: boolean.

    >>> convert_idioms("지금 mp3 파일을 다운받고 있어요")
    지금 엠피쓰리 파일을 다운받고 있어요
    """
    rule = "from idioms.txt"
    out = text

    for line in idioms:
        line = line.split("#")[0].strip()
        if "===" in line:
            str1, str2 = line.split("===")
            if not convert_english_to_hangul_phonemes and re.search(r"[A-Za-z]", str1):
                continue
            if not convert_numbers_to_hangul_phonemes and re.search(r"\d", str1):
                continue
            out = re.sub(str1, str2, out)
    gloss(verbose, out, text, rule)

    return out


cmu = cmudict.dict()
rule_id2text = get_rule_id2text()
idioms = get_idioms()
table = parse_table()



from kss import Kss


def test_select_josa():
    select_josa = Kss("select_josa")
    prefix = "철수"
    josa = "은"
    output = select_josa(prefix, josa)
    assert output == "는"


def test_combine_josa():
    combine_josa = Kss("combine_josa")
    prefix = "철수"
    josa = "은"
    output = combine_josa(prefix, josa)
    assert output == "철수는"


if __name__ == '__main__':
    test_select_josa()
    test_combine_josa()



import jamo
import numpy as np

hangul_distance_map = {}
hangul_distance_map.update({lead: 0.5 for lead in jamo.JAMO_LEADS})
hangul_distance_map.update({vowel: 0.5 for vowel in jamo.JAMO_VOWELS})
hangul_distance_map.update({tail: 0.25 for tail in jamo.JAMO_TAILS})


def hangul_levenshtein(
    a,
    b,
    normalize=False,
    decompose=True,
    lowercase=False,
    verbose=False,
    insertions_cost_ratio=0.6,
    deletions_cost_ratio=0.4,
):
    """
    In most, Korean people can identify a word even if only the lead consonants of the word are given.
    For example, 'ㄱ' and 'ㅅ' can be identified as '감사' and 'ㅎ' and 'ㅇ' can be identified as '하이'.
    But, if only tail consonants of the word are given, it is very hard to identify the word.
    For example, We can't identify it's '감사' if ['ㅁ', None] is given.

    Let's say we have a source word and two candidate words like the following:

    - source_word = "또또칸"
    - candidate_1 = "똑똑한"
    - candidate_2 = "고소한"

    With original Levenshtein distance, the distance between `source_word` and `candidate_1` is 3
    and the distance between `source_word` and `candidate_2` is also 3. But, with Hangul Levenshtein
    distance, the distance between `source_word` and `candidate_1` can be decreased to 0.7, but the
    distance between `source_word` and `candidate_2` is 1.5. And we all know that '똑똑한' is much
    closer to '또또칸' than '고소한'.
    """
    if isinstance(a, str) is False:
        raise TypeError('First argument is not a string!')
    if isinstance(b, str) is False:
        raise TypeError('Second argument is not a string!')
    if a == '':
        return len(b)
    if b == '':
        return len(a)

    if lowercase:
        a = a.lower()
        b = b.lower()

    if decompose:
        a = jamo.h2j(a)
        b = jamo.h2j(b)

    default_distance = 1

    n = len(a)
    m = len(b)
    lev = np.zeros((n + 1, m + 1))

    for i in range(n + 1):
        lev[i, 0] = i
    for i in range(m + 1):
        lev[0, i] = i

    decay_by_length = 1.0
    for i in range(1, n + 1):
        for j in range(1, m + 1):
            # 사전에서 글자에 대한 거리 가져오기, 없으면 기본 거리 사용.
            insertion_cost = hangul_distance_map.get(a[i - 1], default_distance) * decay_by_length
            deletion_cost = hangul_distance_map.get(b[j - 1], default_distance) * decay_by_length

            insertion = lev[i - 1, j] + insertion_cost
            deletion = lev[i, j - 1] + deletion_cost

            if a[i - 1] == b[j - 1]:
                substitution = lev[i - 1, j - 1]
            else:
                substitution_cost = insertion_cost * insertions_cost_ratio + deletion_cost * deletions_cost_ratio
                substitution = lev[i - 1, j - 1] + substitution_cost

            lev[i, j] = min(insertion, deletion, substitution)

        # 글자가 길어질수록 거리를 아주 조금씩 감소시킴.
        # 앞글자를 맞추는게 더 중요하기 때문임.
        decay_by_length -= 0.000001

    if verbose:
        print("levenshtein matrix: {}".format(lev))

    if normalize:
        return lev[n, m] / (n + m / 2)
    else:
        return lev[n, m]



import re

from sentence_split import run_evaluate


def word_splitter(text):
    return text.split(" ")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("datasets", nargs="+")
    parser.add_argument("--write_result")
    parser.add_argument("--write_err")
    args = parser.parse_args()

    word_splitter("foo-bar")  # warm-up

    for dataset in args.datasets:
        run_evaluate(
            dataset,
            lambda text: word_splitter(text),
            args.write_result,
            args.write_err,
        )



# This code is copied from soynlp [https://github.com/lovit/soynlp]
# And modified by Hyunwoong Ko [https://github.com/hyuwoongko]

import re
import sys
from functools import partial
from typing import Union, Tuple, Dict, Any, List

from kss._modules.preprocessing.completed_form import _incompleted_form_ratio
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_type, _check_num_workers, _check_text

hangul_pattern = re.compile(r'[ㄱ-ㅎㅏ-ㅣ가-힣]')
symbols_pattern = re.compile("#+")
punctuations = {'.', '!', '?', '。'}
parenthesis = {'(', ')', '[', ']', '{', '}', '<', '>'}
ellipsis_marks = {"...", "[...]", "(...)", "…", "[…]"}


def _get_ngrams(input_list, n):
    # Fast function to return n-grams from a list of tokens.
    return [item for item in zip(*[input_list[i:] for i in range(n)])]


def _repeating_top_ngram_score(text, n):
    split_text = text.strip().split()
    ngrams = _get_ngrams(split_text, n)
    unique_ngrams = set(ngrams)

    # Find the most frequent ngram in the zipped ngram list
    counts = {
        ngram: {"freq": 0, "num_chars": sum(len(word) for word in ngram)}
        for ngram in unique_ngrams
    }

    for ngram in ngrams:
        counts[ngram]["freq"] += 1
    most_frqnt_ngram = " ".join(max(counts, key=lambda x: counts[x]["freq"]))

    # Find the number of characters the most frequent ngram
    # contributes to the document
    nchar = len(text)
    len_diff = nchar - len(text.replace(most_frqnt_ngram, ""))
    if nchar > 0:
        score = len_diff / nchar
    else:
        # Remove if the document is empty
        score = 1.0
    return score


def _repeating_duplicated_ngrams_score(text, n):
    split_text = text.strip().split()
    ngrams = _get_ngrams(split_text, n)

    counts = {}
    duplicated_nchar = 0
    overlapping_ngrams = 0

    for ngram in ngrams:
        counts[ngram] = counts.get(ngram, 0) + 1
        if counts[ngram] > 1:
            # Count the number of characters in this ngram that haven't been counted already
            duplicated_ngrams = sum(
                len(gram) for gram in ngram[overlapping_ngrams:]
            )
            # Count the spaces between the ngrams
            nspaces = min(n - overlapping_ngrams, n - 1)
            duplicated_nchar += duplicated_ngrams + nspaces
            overlapping_ngrams = n
        overlapping_ngrams = max(overlapping_ngrams - 1, 0)

    nchar = len(text)
    if nchar > 0:
        score = duplicated_nchar / nchar
    else:
        # Remove if the document is empty
        score = 1.0
    return score


def _num_line_by_repeats(text, split_delimiter='\n'):
    split_lines_by_char = [l.strip() for l in text.split(split_delimiter) if len(l.strip()) > 0]
    counter_dict = {}
    for c in split_lines_by_char:
        counter_dict[c] = counter_dict.get(c, 0) + 1

    return max(counter_dict.values())


def _num_line_by_char_repeats(text, split_delimiter='\n'):
    split_lines_by_char = [l.strip()[0] for l in text.split(split_delimiter) if len(l.strip()) > 0]
    counter_dict = {}
    for c in split_lines_by_char:
        counter_dict[c] = counter_dict.get(c, 0) + 1
    return max(counter_dict.values())


def _lines_started_with_bullets_ratio(text):
    lines = [x.strip() for x in text.split('\n') if len(x.strip()) > 0]
    calculated_ratio = sum([x[0] in "•‣⁃⁌⁍∙○●◘◦⦾⦿" for x in lines])
    return calculated_ratio / len(lines)


def _lines_ends_with_ellipsis_ratio(text):
    lines = [x.strip() for x in text.split('\n') if len(x.strip()) > 0]
    num_lines_ends_with_ellipsis = 0
    for line in lines:
        is_lines_ends_with_ellipsis = any([line.endswith(e) for e in ellipsis_marks])
        if is_lines_ends_with_ellipsis:
            num_lines_ends_with_ellipsis += 1
    return num_lines_ends_with_ellipsis / len(lines)


def _symbols_to_word_ratio(text):
    num_symbol_words = 0
    words = text.strip().split()
    for word in words:
        word = word.strip()
        symbol_ratio = len(symbols_pattern.findall(word)) / len(word)
        if word in ellipsis_marks or symbol_ratio > 0.5:
            num_symbol_words += 1
    return num_symbol_words / len(words)


def filter_out(
    text: Union[str, List[str], Tuple[str]],
    min_length: int = 0,
    max_length: int = sys.maxsize,
    min_mean_words_length: int = 0,
    max_mean_words_length: int = sys.maxsize,
    min_words: int = 0,
    max_words: int = sys.maxsize,
    min_lines: int = 0,
    max_lines: int = sys.maxsize,
    min_paragraphs: int = 0,
    max_paragraphs: int = sys.maxsize,
    min_alphabet_ratio: float = 0,
    max_alphabet_ratio: float = 1,
    min_alphanumeric_ratio: float = 0,
    max_alphanumeric_ratio: float = 1,
    min_number_ratio: float = 0,
    max_number_ratio: float = 1,
    min_punctuation_ratio: float = 0,
    max_punctuation_ratio: float = 1,
    min_symbols_to_words_ratio: float = 0,
    max_symbols_to_words_ratio: float = 1,
    min_lines_started_with_bullets_ratio: float = 0,
    max_lines_started_with_bullets_ratio: float = 1,
    min_whitespace_ratio: float = 0,
    max_whitespace_ratio: float = 1,
    min_parenthesis_ratio: float = 0,
    max_parenthesis_ratio: float = 1,
    min_ellipsis_ratio: float = 0,
    max_ellipsis_ratio: float = 1,
    min_hangul_ratio: float = 0,
    max_hangul_ratio: float = 1,
    max_words_length: int = sys.maxsize,
    max_line_repeats: int = sys.maxsize,
    max_line_by_char_repeats: int = sys.maxsize,
    max_paragraph_repeats: int = sys.maxsize,
    max_paragraph_by_char_repeats: int = sys.maxsize,
    max_repeating_top_ngram_repeats_score: float = sys.maxsize,
    max_repeating_duplicate_ngrams_score: float = sys.maxsize,
    ngram_size_for_repeating_top_ngram_repeats: int = 3,
    ngram_size_for_repeating_duplicate_ngrams: int = 3,
    max_hangul_incompleted_form_ratio: float = 1,
    num_workers: Union[int, str] = "auto",
) -> Union[Tuple[bool, Dict[str, Any]], List[Tuple[bool, Dict[str, Any]]]]:
    """
    This filters out bad text based on various conditions.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        min_length (int): minimum length of text
        max_length (int): maximum length of text
        min_mean_words_length (int): minimum mean words length
        max_mean_words_length (int): maximum mean words length
        min_words (int): minimum number of words
        max_words (int): maximum number of words
        min_lines (int): minimum number of lines
        max_lines (int): maximum number of lines
        min_paragraphs (int): minimum number of paragraphs
        max_paragraphs (int): maximum number of paragraphs
        min_alphabet_ratio (float): minimum alphabet ratio
        max_alphabet_ratio (float): maximum alphabet ratio
        min_alphanumeric_ratio (float): minimum alphanumeric ratio
        max_alphanumeric_ratio (float): maximum alphanumeric ratio
        min_number_ratio (float): minimum number ratio
        max_number_ratio (float): maximum number ratio
        min_punctuation_ratio (float): minimum punctuation ratio
        max_punctuation_ratio (float): maximum punctuation ratio
        min_symbols_to_words_ratio (float): minimum symbols to words ratio
        max_symbols_to_words_ratio (float): maximum symbols to words ratio
        min_lines_started_with_bullets_ratio (float): minimum lines started with bullets ratio
        max_lines_started_with_bullets_ratio (float): maximum lines started with bullets ratio
        min_whitespace_ratio (float): minimum whitespace ratio
        max_whitespace_ratio (float): maximum whitespace ratio
        min_parenthesis_ratio (float): minimum parenthesis ratio
        max_parenthesis_ratio (float): maximum parenthesis ratio
        min_ellipsis_ratio (float): minimum ellipsis ratio
        max_ellipsis_ratio (float): maximum ellipsis ratio
        min_hangul_ratio (float): minimum hangul ratio
        max_hangul_ratio (float): maximum hangul ratio
        max_words_length (int): maximum words length
        max_line_repeats (int): maximum line repeats
        max_line_by_char_repeats (int): maximum line by char repeats
        max_paragraph_repeats (int): maximum paragraph repeats
        max_paragraph_by_char_repeats (int): maximum paragraph by char repeats
        max_repeating_top_ngram_repeats_score (float): maximum repeating top ngram repeats score
        max_repeating_duplicate_ngrams_score (float): maximum repeating duplicate ngrams score
        ngram_size_for_repeating_top_ngram_repeats (int): ngram size for repeating top ngram repeats
        ngram_size_for_repeating_duplicate_ngrams (int): ngram size for repeating duplicate ngrams
        max_hangul_incompleted_form_ratio (float): maximum hangul non completed form ratio
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[Tuple[bool, Dict[str, Any]], List[Tuple[bool, Dict[str, Any]]]]: filtered out text or list of filtered out texts

    Examples:
        >>> from kss import Kss
        >>> filter_out = Kss("filter_out")
        >>> text = "▲ 12월 17일 (목)=============================================================================시간 경기내용 방송사=============================================================================[축구] 프랑스 리그 103:00 (AS모나코-스타드 렌) SBS스포츠[축구] 09-10 UEFA 유로파리그04:00 (스파르타프라하-FC코펜하겐) MBC-ESPN[축구] 09-10 프리미어리그05:00 (토트넘-맨체스터시티) SBS스포츠-----------------------------------------------------------------------------[농구] 2009-10 NBA10:00 (LA레이커스-밀워키) SBS스포츠[농구] 2009-10 신한은행 여자농구17:00 (삼성생명-금호생명) SBS스포츠[농구] 2009-10 KCC 프로농구19:00 (KCC-KT) MBC-ESPN19:00 (LG-SK) SBS스포츠-----------------------------------------------------------------------------[배구] 2009-10 NH농협 V리그17:00 (흥국생명-현대건설)19:00 (대한항공-신협상무) KBS N 스포츠-----------------------------------------------------------------------------19:20 [핸드볼] 2009 세계여자선수권대회-----------------------------------------------------------------------------19:00 [배드민턴] 한국 최강전=============================================================================※ 상기 경기일정 및 방송 편성정보는 사정에 따라 변동될 수 있습니다< pre >/한국아이닷컴 뉴스부 한국아이닷컴 뉴스부 '스타화보 VM' 무료다운받기 [**8253+NATE 또는 통화] [ⓒ 인터넷한국일보(www.hankooki.com), 무단 전재 및 재배포 금지]"
        >>> output = filter_out(text, min_mean_words_length=2, max_mean_words_length=10)
        >>> print(output)
        (True, {'reason': 'mean_words_length', 'value': 13.025316455696203})
    """
    text, finish = _check_text(text)

    if finish:
        return True, {"reason": "unsupported_text_type", "value": text}

    min_length = _check_type(min_length, "min_length", int)
    max_length = _check_type(max_length, "max_length", int)
    min_mean_words_length = _check_type(min_mean_words_length, "min_mean_words_length", int)
    max_mean_words_length = _check_type(max_mean_words_length, "max_mean_words_length", int)
    min_words = _check_type(min_words, "min_words", int)
    max_words = _check_type(max_words, "max_words", int)
    min_lines = _check_type(min_lines, "min_lines", int)
    max_lines = _check_type(max_lines, "max_lines", int)
    min_paragraphs = _check_type(min_paragraphs, "min_paragraphs", int)
    max_paragraphs = _check_type(max_paragraphs, "max_paragraphs", int)
    min_alphabet_ratio = _check_type(min_alphabet_ratio, "min_alphabet_ratio", float)
    max_alphabet_ratio = _check_type(max_alphabet_ratio, "max_alphabet_ratio", float)
    min_alphanumeric_ratio = _check_type(min_alphanumeric_ratio, "min_alphanumeric_ratio", float)
    max_alphanumeric_ratio = _check_type(max_alphanumeric_ratio, "max_alphanumeric_ratio", float)
    min_number_ratio = _check_type(min_number_ratio, "min_number_ratio", float)
    max_number_ratio = _check_type(max_number_ratio, "max_number_ratio", float)
    min_punctuation_ratio = _check_type(min_punctuation_ratio, "min_punctuation_ratio", float)
    max_punctuation_ratio = _check_type(max_punctuation_ratio, "max_punctuation_ratio", float)
    min_symbols_to_words_ratio = _check_type(min_symbols_to_words_ratio, "min_symbols_to_words_ratio", float)
    max_symbols_to_words_ratio = _check_type(max_symbols_to_words_ratio, "max_symbols_to_words_ratio", float)
    min_lines_started_with_bullets_ratio = _check_type(min_lines_started_with_bullets_ratio,
                                                       "min_lines_started_with_bullets_ratio", float)
    max_lines_started_with_bullets_ratio = _check_type(max_lines_started_with_bullets_ratio,
                                                       "max_lines_started_with_bullets_ratio", float)
    min_whitespace_ratio = _check_type(min_whitespace_ratio, "min_whitespace_ratio", float)
    max_whitespace_ratio = _check_type(max_whitespace_ratio, "max_whitespace_ratio", float)
    min_parenthesis_ratio = _check_type(min_parenthesis_ratio, "min_parenthesis_ratio", float)
    max_parenthesis_ratio = _check_type(max_parenthesis_ratio, "max_parenthesis_ratio", float)
    min_ellipsis_ratio = _check_type(min_ellipsis_ratio, "min_ellipsis_ratio", float)
    max_ellipsis_ratio = _check_type(max_ellipsis_ratio, "max_ellipsis_ratio", float)
    min_hangul_ratio = _check_type(min_hangul_ratio, "min_hangul_ratio", float)
    max_hangul_ratio = _check_type(max_hangul_ratio, "max_hangul_ratio", float)
    max_words_length = _check_type(max_words_length, "max_words_length", int)
    max_line_repeats = _check_type(max_line_repeats, "max_line_repeats", int)
    max_line_by_char_repeats = _check_type(max_line_by_char_repeats, "max_line_by_char_repeats", int)
    max_paragraph_repeats = _check_type(max_paragraph_repeats, "max_paragraph_repeats", int)
    max_paragraph_by_char_repeats = _check_type(max_paragraph_by_char_repeats, "max_paragraph_by_char_repeats", int)
    max_repeating_top_ngram_repeats_score = _check_type(max_repeating_top_ngram_repeats_score,
                                                        "max_repeating_top_ngram_repeats_score", int)
    max_repeating_duplicate_ngrams_score = _check_type(max_repeating_duplicate_ngrams_score,
                                                       "max_repeating_duplicate_ngrams_score", int)
    ngram_size_for_repeating_top_ngram_repeats = _check_type(ngram_size_for_repeating_top_ngram_repeats,
                                                             "ngram_size_for_repeating_top_ngram_repeats", int)
    ngram_size_for_repeating_duplicate_ngrams = _check_type(ngram_size_for_repeating_duplicate_ngrams,
                                                            "ngram_size_for_repeating_duplicate_ngrams", int)
    max_hangul_incompleted_form_ratio = _check_type(max_hangul_incompleted_form_ratio,
                                                      "max_hangul_incompleted_form_ratio", float)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(
            _filter_out,
            min_length=min_length,
            max_length=max_length,
            min_mean_words_length=min_mean_words_length,
            max_mean_words_length=max_mean_words_length,
            min_words=min_words,
            max_words=max_words,
            min_lines=min_lines,
            max_lines=max_lines,
            min_paragraphs=min_paragraphs,
            max_paragraphs=max_paragraphs,
            min_alphabet_ratio=min_alphabet_ratio,
            max_alphabet_ratio=max_alphabet_ratio,
            min_alphanumeric_ratio=min_alphanumeric_ratio,
            max_alphanumeric_ratio=max_alphanumeric_ratio,
            min_number_ratio=min_number_ratio,
            max_number_ratio=max_number_ratio,
            min_punctuation_ratio=min_punctuation_ratio,
            max_punctuation_ratio=max_punctuation_ratio,
            min_symbols_to_words_ratio=min_symbols_to_words_ratio,
            max_symbols_to_words_ratio=max_symbols_to_words_ratio,
            min_lines_started_with_bullets_ratio=min_lines_started_with_bullets_ratio,
            max_lines_started_with_bullets_ratio=max_lines_started_with_bullets_ratio,
            min_whitespace_ratio=min_whitespace_ratio,
            max_whitespace_ratio=max_whitespace_ratio,
            min_parenthesis_ratio=min_parenthesis_ratio,
            max_parenthesis_ratio=max_parenthesis_ratio,
            min_ellipsis_ratio=min_ellipsis_ratio,
            max_ellipsis_ratio=max_ellipsis_ratio,
            min_hangul_ratio=min_hangul_ratio,
            max_hangul_ratio=max_hangul_ratio,
            max_words_length=max_words_length,
            max_line_repeats=max_line_repeats,
            max_line_by_char_repeats=max_line_by_char_repeats,
            max_paragraph_repeats=max_paragraph_repeats,
            max_paragraph_by_char_repeats=max_paragraph_by_char_repeats,
            max_repeating_top_ngram_repeats_score=max_repeating_top_ngram_repeats_score,
            max_repeating_duplicate_ngrams_score=max_repeating_duplicate_ngrams_score,
            ngram_size_for_repeating_top_ngram_repeats=ngram_size_for_repeating_top_ngram_repeats,
            ngram_size_for_repeating_duplicate_ngrams=ngram_size_for_repeating_duplicate_ngrams,
            max_hangul_incompleted_form_ratio=max_hangul_incompleted_form_ratio,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _filter_out(
    text,
    min_length=0,
    max_length=sys.maxsize,
    min_mean_words_length=0,
    max_mean_words_length=sys.maxsize,
    min_words=0,
    max_words=sys.maxsize,
    min_lines=0,
    max_lines=sys.maxsize,
    min_paragraphs=0,
    max_paragraphs=sys.maxsize,
    min_alphabet_ratio=0,
    max_alphabet_ratio=1,
    min_alphanumeric_ratio=0,
    max_alphanumeric_ratio=1,
    min_number_ratio=0,
    max_number_ratio=1,
    min_punctuation_ratio=0,
    max_punctuation_ratio=1,
    min_symbols_to_words_ratio=0,
    max_symbols_to_words_ratio=1,
    min_lines_started_with_bullets_ratio=0,
    max_lines_started_with_bullets_ratio=1,
    min_whitespace_ratio=0,
    max_whitespace_ratio=1,
    min_parenthesis_ratio=0,
    max_parenthesis_ratio=1,
    min_ellipsis_ratio=0,
    max_ellipsis_ratio=1,
    min_hangul_ratio=0,
    max_hangul_ratio=1,
    max_words_length=sys.maxsize,
    max_line_repeats=sys.maxsize,
    max_line_by_char_repeats=sys.maxsize,
    max_paragraph_repeats=sys.maxsize,
    max_paragraph_by_char_repeats=sys.maxsize,
    max_repeating_top_ngram_repeats_score=sys.maxsize,
    max_repeating_duplicate_ngrams_score=sys.maxsize,
    ngram_size_for_repeating_top_ngram_repeats=3,
    ngram_size_for_repeating_duplicate_ngrams=3,
    max_hangul_incompleted_form_ratio=1,
):
    words = text.split()
    lines = [x.strip() for x in text.split('\n') if len(x.strip()) > 0]
    paragraphs = [x.strip() for x in text.split('\n\n') if len(x.strip()) > 0]
    eps = 1e-6

    # min/max
    len_text = len(text)
    if not min_length <= len_text <= max_length:
        return True, {"reason": "length", "value": len_text}

    len_text += eps
    mean_words_length = sum(len(w) for w in words) / len(words)
    if not min_mean_words_length <= mean_words_length <= max_mean_words_length:
        return True, {"reason": "mean_words_length", "value": mean_words_length}

    len_words = len(words)
    if not min_words <= len_words <= max_words:
        return True, {"reason": "num_words", "value": len_words}

    len_lines = len(lines)
    if not min_lines <= len_lines <= max_lines:
        return True, {"reason": "num_lines", "value": len_lines}

    len_paragraphs = len(paragraphs)
    if not min_paragraphs <= len_paragraphs <= max_paragraphs:
        return True, {"reason": "num_paragraphs", "value": len_paragraphs}

    alphabet_ratio = sum(c.isalpha() for c in text) / len_text
    if not min_alphabet_ratio <= alphabet_ratio <= max_alphabet_ratio:
        return True, {"reason": "alphabet_ratio", "value": alphabet_ratio}

    alphanumeric_ratio = sum(c.isalnum() for c in text) / len_text
    if not min_alphanumeric_ratio <= alphanumeric_ratio <= max_alphanumeric_ratio:
        return True, {"reason": "alphanumeric_ratio", "value": alphanumeric_ratio}

    digits_ratio = sum(c.isdigit() for c in text) / len_text
    if not min_number_ratio <= digits_ratio <= max_number_ratio:
        return True, {"reason": "number_ratio", "value": digits_ratio}

    punctuation_ratio = sum(c in punctuations for c in text) / len_text
    if not min_punctuation_ratio <= punctuation_ratio <= max_punctuation_ratio:
        return True, {"reason": "punctuation_ratio", "value": punctuation_ratio}

    symbols_to_word_ratio = _symbols_to_word_ratio(text)
    if not min_symbols_to_words_ratio <= symbols_to_word_ratio <= max_symbols_to_words_ratio:
        return True, {"reason": "symbols_to_words_ratio", "value": symbols_to_word_ratio}

    lines_started_with_bullets_ratio = _lines_started_with_bullets_ratio(text)
    if not min_lines_started_with_bullets_ratio <= lines_started_with_bullets_ratio <= max_lines_started_with_bullets_ratio:
        return True, {"reason": "lines_started_with_bullets_ratio", "value": lines_started_with_bullets_ratio}

    whitespace_ratio = sum(c.isspace() for c in text) / len_text
    if not min_whitespace_ratio <= whitespace_ratio <= max_whitespace_ratio:
        return True, {"reason": "whitespace_ratio", "value": whitespace_ratio}

    parenthesis_ratio = sum(c in parenthesis for c in text) / len_text
    if not min_parenthesis_ratio <= parenthesis_ratio <= max_parenthesis_ratio:
        return True, {"reason": "parenthesis_ratio", "value": parenthesis_ratio}

    ellipsis_ratio = _lines_ends_with_ellipsis_ratio(text)
    if not min_ellipsis_ratio <= ellipsis_ratio <= max_ellipsis_ratio:
        return True, {"reason": "ellipsis_ratio", "value": ellipsis_ratio}

    hangul_ratio = len(hangul_pattern.findall(text)) / len_text
    if not min_hangul_ratio <= hangul_ratio <= max_hangul_ratio:
        return True, {"reason": "hangul_ratio", "value": hangul_ratio}

    _max_words_length = max(len(w) for w in words)
    if max_words_length <= _max_words_length:
        return True, {"reason": "max_words_length", "value": _max_words_length}

    _max_line_repeats = _num_line_by_repeats(text)
    if max_line_repeats <= _max_line_repeats:
        return True, {"reason": "line_repeats", "value": _max_line_repeats}

    _max_line_by_char_repeats = _num_line_by_char_repeats(text)
    if max_line_by_char_repeats <= _max_line_by_char_repeats:
        return True, {"reason": "line_by_char_repeats", "value": _max_line_by_char_repeats}

    _max_paragraph_repeats = _num_line_by_char_repeats(text, split_delimiter='\n\n')
    if max_paragraph_repeats <= _max_paragraph_repeats:
        return True, {"reason": "paragraph_repeats", "value": _max_paragraph_repeats}

    _max_paragraph_by_char_repeats = _num_line_by_char_repeats(text, split_delimiter='\n\n')
    if max_paragraph_by_char_repeats <= _max_paragraph_by_char_repeats:
        return True, {"reason": "paragraph_by_char_repeats", "value": _max_paragraph_by_char_repeats}

    _max_repeating_top_ngram_repeats_score = _repeating_top_ngram_score(text,
                                                                        ngram_size_for_repeating_top_ngram_repeats)
    if max_repeating_top_ngram_repeats_score <= _max_repeating_top_ngram_repeats_score:
        return True, {"reason": "repeating_top_ngram_repeats_score", "value": _max_repeating_top_ngram_repeats_score}

    _max_repeating_duplicate_ngrams_score = _repeating_duplicated_ngrams_score(text,
                                                                               ngram_size_for_repeating_duplicate_ngrams)
    if max_repeating_duplicate_ngrams_score <= _max_repeating_duplicate_ngrams_score:
        return True, {"reason": "repeating_duplicate_ngrams_score", "value": _max_repeating_duplicate_ngrams_score}

    _hangul_incompleted_form_ratio = _incompleted_form_ratio(text)
    if max_hangul_incompleted_form_ratio <= _hangul_incompleted_form_ratio:
        return True, {"reason": "hangul_incompleted_form_ratio", "value": _hangul_incompleted_form_ratio}

    return False, {"reason": "pass", "value": None}



# -*- coding: utf-8 -*-
"""
    hangulize.processing
    ~~~~~~~~~~~~~~~~~~~~

    :copyright: (c) 2010-2017 by Heungsub Lee
    :license: BSD, see LICENSE for more details.
"""
from __future__ import absolute_import

from kss._modules.hangulization.hangulize.hangul import EU, Null, NG, UnicodeHangulError, join, split
from kss._modules.hangulization.hangulize.models import Choseong, Impurity, Jongseong, Jungseong


__all__ = ['complete_syllable', 'complete_syllables', 'split_phonemes',
           'join_phonemes']


def complete_syllable(syllable):
    """Inserts the default jungseong or jongseong if it is not exists::

        >>> complete_syllable((Jungseong(YO),))
        (u'ㅇ', u'ㅛ', u'')
        >>> print hangulize.hangul.join(_)
        요

    """
    syllable = list(syllable)
    components = [type(ph) for ph in syllable]
    if Choseong not in components:
        syllable.insert(0, Choseong(NG))
    if Jungseong not in components:
        syllable.insert(1, Jungseong(EU))
    if Jongseong not in components:
        syllable.insert(2, Jungseong(Null))
    return tuple((ph.letter for ph in syllable))


def complete_syllables(phonemes):
    """Separates each syllables and completes every syllable."""
    components, syllable = [Choseong, Jungseong, Jongseong], []
    if phonemes:
        for ph in phonemes:
            comp = type(ph)
            new_syllable = (comp is Impurity or syllable and
                            components.index(comp) <=
                            components.index(type(syllable[-1])))
            if new_syllable:
                if syllable:
                    yield complete_syllable(syllable)
                    syllable = []
                if comp is Impurity:
                    yield (ph,)
                    continue
            syllable.append(ph)
        if syllable:
            yield complete_syllable(syllable)


def split_phonemes(word):
    """Returns the splitted phonemes from the word.

        >>> split_phonemes(u'안녕') #doctest: +NORMALIZE_WHITESPACE
        (<Choseong 'ㅇ'>, <Jungseong 'ㅏ'>, <Jongseong 'ㄴ'>,
         <Choseong 'ㄴ'>, <Jungseong 'ㅕ'>, <Jongseong 'ㅇ'>)
    """
    result = []
    for c in word:
        try:
            c = split(c)
            result.append(Choseong(c[0]))
            result.append(Jungseong(c[1]))
            if c[2] is not Null:
                result.append(Jongseong(c[2]))
        except UnicodeHangulError:
            result.append(Impurity(c))
    return tuple(result)


def join_phonemes(phonemes):
    """Returns the word from the splitted phonemes::

        >>> print join_phonemes((Jungseong(A), Jongseong(N),
        ...                      Choseong(N), Jungseong(YEO), Jongseong(NG)))
        안녕

    """
    syllables = complete_syllables(phonemes)
    chars = (join(syl) for syl in syllables)
    return reduce(unicode.__add__, chars)



import math
import re

ENGLISH = "rRseEfaqQtTdwWczxvgASDFGZXCVkoiOjpuPhynbmlYUIHJKLBNM"
HANGUL = "ㄱㄲㄴㄷㄸㄹㅁㅂㅃㅅㅆㅇㅈㅉㅊㅋㅌㅍㅎㅁㄴㅇㄹㅎㅋㅌㅊㅍㅏㅐㅑㅒㅓㅔㅕㅖㅗㅛㅜㅠㅡㅣㅛㅕㅑㅗㅓㅏㅣㅠㅜㅡ"
LEADS = "ㄱㄲㄴㄷㄸㄹㅁㅂㅃㅅㅆㅇㅈㅉㅊㅋㅌㅍㅎ"
VOWELS = "ㅏㅐㅑㅒㅓㅔㅕㅖㅗㅘㅙㅚㅛㅜㅝㅞㅟㅠㅡㅢㅣ"
TAILS = "ㄱㄲㄳㄴㄵㄶㄷㄹㄺㄻㄼㄽㄾㄿㅀㅁㅂㅄㅅㅆㅇㅈㅊㅋㅌㅍㅎ"

HANGUL_FIRST = 44032
HANGUL_LAST = 55203
VOWELS_FIRST = 28
CONSONANT_FIRST = 12593
VOWEL_LAST = 12643

ENGLISH_INDEX = (lambda en: {en[i]: i for i in range(len(en))})(ENGLISH)
HANGUL_INDEX = (lambda kr: {i: w for i, w in enumerate(kr)})(HANGUL)

CONNECTABLE_CONSONANTS = {
    'ㄱㅅ': 'ㄳ',
    'ㄴㅈ': 'ㄵ',
    'ㄴㅎ': 'ㄶ',
    'ㄹㄱ': 'ㄺ',
    'ㄹㅁ': 'ㄻ',
    'ㄹㅂ': 'ㄼ',
    'ㄹㅅ': 'ㄽ',
    'ㄹㅌ': 'ㄾ',
    'ㄹㅍ': 'ㄿ',
    'ㄹㅎ': 'ㅀ',
    'ㅂㅅ': 'ㅄ'
}

CONNECTABLE_VOWELS = {
    'ㅗㅏ': 'ㅘ',
    'ㅗㅐ': 'ㅙ',
    'ㅗㅣ': 'ㅚ',
    'ㅜㅓ': 'ㅝ',
    'ㅜㅔ': 'ㅞ',
    'ㅜㅣ': 'ㅟ',
    'ㅡㅣ': 'ㅢ'
}


def is_vowel(e):
    return [k for k, v in HANGUL_INDEX.items() if v == e][0] >= VOWELS_FIRST


def is_hangul(char):
    if len(char) > 1:
        raise Exception("`char` must be a single character.")
    return re.match("[ㄱ-ㅎㅏ-ㅣ가-힣]", char) is not None


def index_of(val, _list):
    try:
        return _list.index(val)
    except ValueError:
        return -1


def generate_hangul(*args):
    return chr(44032 + args[0][0] * 588 + args[0][1] * 28 + args[0][2] + 1)


def separate_hangul(hangul_text):
    code = ord(hangul_text)

    if HANGUL_FIRST <= code <= HANGUL_LAST:
        lead = math.floor((code - HANGUL_FIRST) / 588)
        vowel = math.floor((code - HANGUL_FIRST - lead * 588) / 28)
        tail = code - HANGUL_FIRST - lead * 588 - vowel * 28 - 1
        vowel_1, vowel_2, tail_1, tail_2 = vowel, -1, tail, -1

        if vowel == index_of("ㅘ", VOWELS):
            vowel_1, vowel_2 = index_of("ㅗ", HANGUL), index_of("ㅏ", HANGUL)
        elif vowel == index_of("ㅙ", VOWELS):
            vowel_1, vowel_2 = index_of("ㅗ", HANGUL), index_of("ㅐ", HANGUL)
        elif vowel == index_of("ㅚ", VOWELS):
            vowel_1, vowel_2 = index_of("ㅗ", HANGUL), index_of("ㅣ", HANGUL)
        elif vowel == index_of("ㅝ", VOWELS):
            vowel_1, vowel_2 = index_of("ㅜ", HANGUL), index_of("ㅓ", HANGUL)
        elif vowel == index_of("ㅞ", VOWELS):
            vowel_1, vowel_2 = index_of("ㅜ", HANGUL), index_of("ㅔ", HANGUL)
        elif vowel == index_of("ㅟ", VOWELS):
            vowel_1, vowel_2 = index_of("ㅜ", HANGUL), index_of("ㅣ", HANGUL)
        elif vowel == index_of("ㅢ", VOWELS):
            vowel_1, vowel_2 = index_of("ㅡ", HANGUL), index_of("ㅣ", HANGUL)

        if tail == index_of("ㄳ", TAILS):
            tail_1, tail_2 = index_of("ㄱ", HANGUL), index_of("ㅅ", HANGUL)
        elif tail == index_of("ㄵ", TAILS):
            tail_1, tail_2 = index_of("ㄴ", HANGUL), index_of("ㅈ", HANGUL)
        elif tail == index_of("ㄶ", TAILS):
            tail_1, tail_2 = index_of("ㄴ", HANGUL), index_of("ㅎ", HANGUL)
        elif tail == index_of("ㄺ", TAILS):
            tail_1, tail_2 = index_of("ㄹ", HANGUL), index_of("ㄱ", HANGUL)
        elif tail == index_of("ㄻ", TAILS):
            tail_1, tail_2 = index_of("ㄹ", HANGUL), index_of("ㅁ", HANGUL)
        elif tail == index_of("ㄼ", TAILS):
            tail_1, tail_2 = index_of("ㄹ", HANGUL), index_of("ㅂ", HANGUL)
        elif tail == index_of("ㄽ", TAILS):
            tail_1, tail_2 = index_of("ㄹ", HANGUL), index_of("ㅅ", HANGUL)
        elif tail == index_of("ㄾ", TAILS):
            tail_1, tail_2 = index_of("ㄹ", HANGUL), index_of("ㅌ", HANGUL)
        elif tail == index_of("ㄿ", TAILS):
            tail_1, tail_2 = index_of("ㄹ", HANGUL), index_of("ㅍ", HANGUL)
        elif tail == index_of("ㅀ", TAILS):
            tail_1, tail_2 = index_of("ㄹ", HANGUL), index_of("ㅎ", HANGUL)
        elif tail == index_of("ㅄ", TAILS):
            tail_1, tail_2 = index_of("ㅂ", HANGUL), index_of("ㅅ", HANGUL)

        if vowel_2 == -1 and vowel != -1:
            vowel_1 = index_of(VOWELS[vowel], HANGUL)
        if tail_2 == -1 and tail != -1:
            tail_1 = index_of(TAILS[tail], HANGUL)

        return [lead, vowel_1, vowel_2, tail_1, tail_2]

    elif CONSONANT_FIRST <= code <= VOWEL_LAST:
        if index_of(hangul_text, LEADS) > -1:
            lead = index_of(hangul_text, HANGUL)
            return [lead, -1, -1, -1, -1]

        elif index_of(hangul_text, VOWELS) > -1:
            vowel = index_of(hangul_text, VOWELS)
            vowel_1, vowel_2 = vowel, -1
            if vowel == index_of("ㅘ", VOWELS):
                vowel_1, vowel_2 = index_of("ㅗ", HANGUL), index_of("ㅏ", HANGUL)
            elif vowel == index_of("ㅙ", VOWELS):
                vowel_1, vowel_2 = index_of("ㅗ", HANGUL), index_of("ㅐ", HANGUL)
            elif vowel == index_of("ㅚ", VOWELS):
                vowel_1, vowel_2 = index_of("ㅗ", HANGUL), index_of("ㅣ", HANGUL)
            elif vowel == index_of("ㅝ", VOWELS):
                vowel_1, vowel_2 = index_of("ㅜ", HANGUL), index_of("ㅓ", HANGUL)
            elif vowel == index_of("ㅞ", VOWELS):
                vowel_1, vowel_2 = index_of("ㅜ", HANGUL), index_of("ㅔ", HANGUL)
            elif vowel == index_of("ㅟ", VOWELS):
                vowel_1, vowel_2 = index_of("ㅜ", HANGUL), index_of("ㅣ", HANGUL)
            elif vowel == index_of("ㅢ", VOWELS):
                vowel_1, vowel_2 = index_of("ㅡ", HANGUL), index_of("ㅣ", HANGUL)

            if vowel_2 == -1:
                vowel_1 = index_of(VOWELS[vowel], HANGUL)

            return [-1, vowel_1, vowel_2, -1, -1]

    return [-1, -1, -1, -1, -1]



# -*- coding: utf-8 -*-
"""
    hangulize.langs
    ~~~~~~~~~~~~~~~

    The languages that Hangulize supports.

    :copyright: (c) 2010-2013 by Heungsub Lee
    :license: BSD, see LICENSE for more details.
"""
import os
import re


p = os.path


def list_langs():
    """Returns the supported language code list."""
    ext = p.extsep + 'py'
    init = '__init__' + ext
    def _list_langs(prefix='', path=None):
        path = path or p.dirname(__file__)
        # helpers
        name = lambda x: prefix + re.sub(re.escape(ext) + '$', '', x)
        def is_lang(x):
            if x.startswith(init):
                return False
            x = p.join(path, x)
            return p.isdir(x) and p.isfile(p.join(x, init)) or \
                   p.isfile(x) and x.endswith(ext)
        # find top-level language modules
        langs = [name(x) for x in os.listdir(path) if is_lang(x)]
        # find sub language modules
        for lang in langs:
            _path = p.join(path, lang)
            if p.isdir(_path):
                langs += _list_langs(prefix=lang + '.', path=_path)
        langs.sort()
        return langs
    return _list_langs()


def get_list():
    """Deprecated with 0.0.6. Use :func:`hangulize.langs.list_langs`
    instead.
    """
    import warnings
    warnings.warn('get_list() has been deprecated, use list_langs() instead',
                  DeprecationWarning)
    return list_langs()



import html
import re
import sys
import unicodedata
from functools import partial
from typing import Union, List, Optional, Tuple

from bs4 import BeautifulSoup

from kss._modules.jamo.utils import _j2h
from kss._modules.preprocessing.half2full import _half2full
from kss._modules.preprocessing.reduce_repeats import _reduce_char_repeats, _reduce_emoticon_repeats
from kss._modules.preprocessing.remove_invisible_chars import _remove_invisible_chars
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_type, _check_num_workers

doubled_spaces_pattern = re.compile('(\s)\\1{2,}')


def normalize(
    text: Union[str, List[str], Tuple[str]],
    normalization_type: Optional[str] = None,
    allow_doubled_spaces: bool = True,
    allow_html_tags: bool = True,
    allow_html_escape: bool = True,
    allow_halfwidth_hangul: bool = True,
    allow_hangul_jamo: bool = True,
    allow_invisible_chars: bool = True,
    reduce_char_repeats_over: int = sys.maxsize,
    reduce_emoticon_repeats_over: int = sys.maxsize,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This normalizes text with various options.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        normalization_type (Optional[str]): normalization type
        allow_doubled_spaces (bool): whether to allow doubled spaces or not
        allow_html_tags (bool): whether to allow HTML tags or not
        allow_html_escape (bool): whether to allow HTML unescape or not
        allow_halfwidth_hangul (bool): whether to allow halfwidth Hangul or not
        allow_hangul_jamo (bool): whether to allow Hangul jamo or not
        allow_invisible_chars (bool): whether to allow invisible characters or not
        reduce_char_repeats_over (int): the maximum number of character that can be repeated
        reduce_emoticon_repeats_over (int): the maximum number of emoticon that can be repeated
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: normalized text or list of normalized texts

    Examples:
        >>> from kss import Kss
        >>> normalize = Kss("normalize")
        >>> text = "안녕\u200b하세요 ﾻﾻﾻﾻﾻﾻ   <br>오늘\u200b은 날이 참 좋네요.\\n\\n\\n200 &lt; 300 &amp; 400"
        >>> normalize(text, allow_doubled_spaces=False, allow_html_tags=False, allow_html_escape=False, allow_halfwidth_hangul=False, allow_hangul_jamo=False, allow_invisible_chars=False, reduce_char_repeats_over=2, reduce_emoticon_repeats_over=2)
        '안녕하세요 ㅋㅋ 오늘은 날이 참 좋네요.\\n200 < 300 & 400'
    """

    text, finish = _check_text(text)

    if finish:
        return text

    if normalization_type is not None:
        normalization_type = _check_type(normalization_type, "normalization_type", str)
    allow_doubled_spaces = _check_type(allow_doubled_spaces, "allow_doubled_spaces", bool)
    allow_html_tags = _check_type(allow_html_tags, "allow_html_tags", bool)
    allow_html_escape = _check_type(allow_html_escape, "allow_html_escape", bool)
    allow_halfwidth_hangul = _check_type(allow_halfwidth_hangul, "allow_halfwidth_hangul", bool)
    allow_hangul_jamo = _check_type(allow_hangul_jamo, "allow_hangul_jamo", bool)
    reduce_char_repeats_over = _check_type(reduce_char_repeats_over, "reduce_char_repeats_over", int)
    reduce_emoticon_repeats_over = _check_type(reduce_emoticon_repeats_over, "reduce_emoticon_repeats_over", int)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(
            _normalize,
            normalization_type=normalization_type,
            allow_doubled_spaces=allow_doubled_spaces,
            allow_html_tags=allow_html_tags,
            allow_html_escape=allow_html_escape,
            allow_halfwidth_hangul=allow_halfwidth_hangul,
            allow_hangul_jamo=allow_hangul_jamo,
            allow_invisible_chars=allow_invisible_chars,
            reduce_char_repeats_over=reduce_char_repeats_over,
            reduce_emoticon_repeats_over=reduce_emoticon_repeats_over,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _normalize(
    text,
    normalization_type=None,
    allow_doubled_spaces=True,
    allow_html_tags=True,
    allow_html_escape=True,
    allow_halfwidth_hangul=True,
    allow_hangul_jamo=True,
    allow_invisible_chars=True,
    reduce_char_repeats_over=sys.maxsize,
    reduce_emoticon_repeats_over=sys.maxsize,
):
    if normalization_type is not None:
        if not isinstance(normalization_type, str):
            raise TypeError("normalization_type should be a string")

        normalization_type = normalization_type.upper()
        if normalization_type not in ['NFC', 'NFKC', 'NFD', 'NFKD']:
            raise ValueError("normalization_type should be one of 'NFC', 'NFKC', 'NFD', 'NFKD'")

        text = unicodedata.normalize(normalization_type, text)

    if not allow_doubled_spaces:
        text = doubled_spaces_pattern.sub('\\1', text)

    if not allow_html_tags:
        text = BeautifulSoup(text, 'html.parser').text

    if not allow_html_escape:
        text = html.unescape(text)

    if not allow_halfwidth_hangul:
        # NFKC and NFKD convert halfwidth Hangul to fullwidth Hangul automatically.
        text = _half2full(text) if normalization_type is None or "K" not in normalization_type else text

    if not allow_hangul_jamo:
        # NFC and NFKC convert Hangul jamo to Hangul compatibility jamo automatically.
        text = _j2h(text) if normalization_type is None or "C" not in normalization_type else text

    if not allow_invisible_chars:
        text = _remove_invisible_chars(text)

    if 0 < reduce_emoticon_repeats_over < sys.maxsize:
        text = _reduce_emoticon_repeats(text, reduce_emoticon_repeats_over)

    if 0 < reduce_char_repeats_over < sys.maxsize:
        text = _reduce_char_repeats(text, reduce_char_repeats_over)

    return text



from typing import Union, List

from kss._modules.josa.utils import _check_text, _check_num_workers, _combine_josa, _run_job, _select_josa


def select_josa(
    prefix: Union[str, List[str]],
    josa: Union[str, List[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This selects the correct josa for the given prefix.

    Args:
        prefix (Union[str, List[str]): single prefix or list of prefixes
        josa (Union[str, List[str]): single josa or list of josas
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: the correct josa for the given prefix

    Examples:
        >>> from kss import Kss
        >>> select_josa = Kss("select_josa")
        >>> prefix = "철수"
        >>> josa = "은"
        >>> output = select_josa(prefix, josa)
        >>> print(output)
        "는"

    References:
        This was copied from [tossi](https://github.com/what-studio/tossi) and modified by Kss
    """
    prefix, josa = _check_text(prefix, josa, "prefix", "josa")
    num_workers = _check_num_workers(prefix, josa, "prefix", "josa", num_workers)

    return _run_job(
        func=_select_josa,
        input_1=prefix,
        input_2=josa,
        input_1_name="prefix",
        input_2_name="josa",
        num_workers=num_workers,
    )


def combine_josa(
    prefix: Union[str, List[str]],
    josa: Union[str, List[str]],
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This combines the given prefix and josa.

    Args:
        prefix (Union[str, List[str]): single prefix or list of prefixes
        josa (Union[str, List[str]): single josa or list of josas
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: the combined prefix and josa

    Examples:
        >>> from kss import Kss
        >>> combine_josa = Kss("combine_josa")
        >>> prefix = "철수"
        >>> josa = "은"
        >>> output = combine_josa(prefix, josa)
        >>> print(output)
        "철수는"

    References:
        This was copied from [tossi](https://github.com/what-studio/tossi) and modified by Kss
    """
    prefix, josa = _check_text(prefix, josa, "prefix", "josa")
    num_workers = _check_num_workers(prefix, josa, "prefix", "josa", num_workers)

    return _run_job(
        func=_combine_josa,
        input_1=prefix,
        input_2=josa,
        input_1_name="prefix",
        input_2_name="josa",
        num_workers=num_workers,
    )



from kss import Kss


def test_augment():
    augment = Kss("augment")
    input_text = """앞서 지난해 11월, 보이저 1호는 명령을 수신하고 수행하는 데엔 문제가 없었지만 통신 장치에 문제가 생겨 과학·엔지니어링 데이터가 지구로 전송되지 않았던 바 있다. 당시 그들은 컴퓨터 시스템을 재시작하고 문제의 근본적인 원인을 파악하기 위해 명령을 보내려고 시도했고, 이달 1일 '포크'라는 명령을 보냈다."""
    output = augment(input_text)
    assert output == "앞서 지난해 11월, 보이저 1호는 명령을 수신하고 시행하는 데엔 문제가 없었지만 송신 장비에 문제가 생겨 과학·엔지니어링 데이터가 지구로 전송되지 않았던 바 있다. 당시 그들은 컴퓨터 시스템을 재시작하고 문제의 근본적인 원인을 파악하기 위해 명령을 보내려고 시도했고, 이달 1일 '포크'라는 명령을 보냈다."

if __name__ == '__main__':
    test_augment()



import re
from functools import partial
from typing import Union, List, Tuple

from kss._utils.logger import highlight_diffs, logger
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_type, _check_num_workers

normalization_open = {
    "< ": "<",
    "〈 ": "<",
    "〈": "<",
    "[ ": "[",
    "( ": "(",
}

normalization_close = {
    " >": ">",
    "〉": ">",
    " 〉": ">",
    " ]": "]",
    " )": ")",
}

normalization_others = {
    "\\n": "\n",
    ". .": "..",
    ". ?": ".?",
    ". !": ".!",
    "! !": "!!",
    "! .": "!.",
    "! ?": "!?",
    "? .": "?.",
    "? !": "?!",
    "? ?": "??",
    ". \" ": ".\" ",
    "! \" ": "!\" ",
    "? \" ": "?\" ",
    "  ": " ",
    "\n ": "\n",
    " \n": "\n",
}

skip_words = [
    "부친상", "모친상", "조부상", "조모상",
    "부고", "별세", "총경 전보", "소방경",
    "인사]",
]

media_list = ['10+Star', 'News1', '21C FA VISION', '365베케이션', '뉴스1', '뉴스핌', '911TV', 'APP저널', 'BBS불교방송', 'BI코리아', 'C3',
              'CARGUY', 'CBS', 'CEO&', 'CEO스코어데일리', 'CIOBIZ+', 'CIOCISO', 'CLO', 'CMC가족오락TV', 'CNB국회방송', 'EBN', 'EPnC',
              'ESCO', 'FAM타임스', 'FA저널', 'FSTV', 'Factiva', 'GIB거제인터넷방송', 'GQ KOREA', 'Groove Korea', 'G밸리', 'HKBC환경방송',
              'HR Insight', 'IP노믹스', 'IT NEWS', 'IT동아', 'IT조선', 'JBC까', 'JTBC', 'K STAR', 'KBS대전', 'KBS미디어', 'KBS춘천',
              'KNN', 'KTV국민방송', 'Kjtimes', 'K모바일', 'MBC경남', 'MBN', 'MBN스타', 'MFG', 'MFIGHT', 'MK스포츠', 'MTN머니투데이방송',
              'Mom대로 키워라', 'M이코노미', 'NBC-1TV', 'NDSOFT', 'NSP통신', 'OBS경인TV', 'OBS플러스', 'OLEDNET', 'OSEN', 'QBS', 'RPM9',
              'RTN부동산TV', 'SBS CNBC', 'SBS Sports', 'SBS funE', 'SEOUL', 'SPACE', 'SPO TV', 'SPOTV NEWS', 'SP투데이',
              'STN SPORTS', 'STYLER', 'TBN한국교통방송', 'TBS교통방송', 'TJB대전방송', 'TK TIMES', 'TV리포트', 'TV조선', 'YTN', 'YTN사이언스',
              'e대학저널', 'e메디코파마', 'e헬스통신', 'icross 교차로', 'jtbc GOLF', '가구저널', '가톨릭평화방송', '건강다이제스트', '건설경제', '건축리포트와이드',
              '건축문화', '게임동아', '게임메카', '게임조선', '게임포커스', '겜툰', '경기시사투데이', '경남매일', '경제풍월', '경제플러스', '경찰방송', '경향게임스',
              '계간 조각', '계장기술', '골닷컴', '골드앤와이즈', '골프가이드', '골프다이제스트', '골프먼스리', '골프세미나', '골프저널', '골프코스 관리정보', '골프포위민',
              '골프한국', '공구사랑', '공구저널', '공무원저널', '공예+디자인', '과학동아', '과학소년', '광주드림', '교육희망', '국방과기술', '국악방송', '국제골프',
              '국토매일', '국토와교통', '국회방송', '군사저널', '귀금속과시계', '그린데일리', '그린투데이', '글로벌이코노믹', '글마루', '금융플러스', '금형기술', '금형저널',
              '기계&자동화', '기술과경영', '기어박스', '까사리빙', '꼬망세', '나라경제', '나우앤티비', '낚시춘추', '난과생활', '내셔널지오그래픽', '내외경제TV',
              '내외전기통신저널', '내외통신', '냉동공조', '네비포유', '네이버 보도자료', '네일홀릭', '넥스트데일리', '넥스트이코노미', '넷퓨', '넷프로', '노동과세계', '노블리안',
              '노트포럼', '녹색경제', '녹색저널', '뉴데일리', '뉴데일리경제', '뉴시스', '뉴트리앤', '다다미디어', '다산저널', '다아라매거진', '다이어트데일리', '닥터W',
              '대덕넷', '대학경제', '대학내일', '대한간호', '대한포커스', '댄스포럼', '더 기어', '더 퍼스트클래스', '더갤러리아', '더게임스', '더골프', '더그아웃', '더닥터',
              '더리더', '더모토', '더뮤지컬', '더바이어', '더바이크', '더불어사는사회', '더셀러브리티', '더스쿠프', '더스포츠', '더아이오토', '더트래블러', '더팩트',
              '더플라워', '데이즈드앤컨퓨즈드', '데이터넷', '데일리 오에스티', '데일리e스포츠', '데일리게임', '데일리경제', '데일리그리드', '데일리대한민국', '데일리로그',
              '데일리머니', '데일리메디', '데일리메디팜', '데일리스포츠한국', '데일리시사닷컴', '데일리시큐', '데일리안', '데일리저널', '데일리중앙', '데일리카', '데일리컬쳐',
              '데일리투머로우', '데일리팜', '데일리펫', '데일리한국', '데일리환경', '데코저널', '덴탈아리랑', '덴탈투데이', '덴탈포커스', '덴포라인', '도시문제', '동아닷컴',
              '동아비즈니스리뷰', '동아사이언스', '동아환경방송', '드림저널', '디바이스마트 매거진', '디스패치', '디아이투데이', '디지털경제', '디지털데일리', '디지털브러시',
              '디지털타임스', '뚜르드몽드', '라움 오', '라이드매거진', '라이브엔', '라펜트', '라포르시안', '랜드데일리', '러브즈뷰티', '럭스맨', '레미콘·아스콘·골재',
              '로드테스트', '로봇기술', '로이슈', '로피시엘옴므', '루어앤플라이피싱', '리빙센스', '리테일매거진', '마니아리포트', '마리끌레르', '마리나스', '마이더스',
              '마이데일리', '마이웨딩', '마이크로소프트웨어', '마케팅투데이', '맘&앙팡', '매거진아트', '매경게임진', '매경닷컴', '매경이코노미', '매경헬스', '매일경제',
              '매일경제TV', '매일투데이', '맥심 코리아', '머니S', '머니위크MNB', '머니투데이', '머니투데이 더벨', '머슬앤맥스큐', '메가오토', '메디칼미디어', '메디칼업저버',
              '메디칼통신', '메디칼트리뷴', '메디컬리포트', '메디컬투데이', '메디컬헤럴드', '메디팜스투데이', '모빌리스타', '모션컨트롤', '모터그래프', '모터리언', '모터매거진',
              '모터스라인', '모터트렌드', '모토야', '무비스트', '무인화기술', '무카스', '문구스타일', '문화관광저널', '문화예술TV21', '문화웹진 나비', '문화저널21',
              '문화투데이', '문화플러스', '문화플러스서울', '미디어광명', '미디어스', '미디어오늘', '미디어펜', '미디어하남', '미래경제', '미래한국', '미술과비평', '미술세계',
              '미즈코치', '미즈호스트', '민중의소리', '바끄로', '바다낚시&SEA LURE', '바스켓코리아', '바앤다이닝', '바이라인 네트워크', '바이오스펙테이터', '바이커즈랩',
              '바이크매거진', '반도체네트워크', '발리볼코리아닷컴', '방송기술저널', '배관기술', '배드민턴매거진', '백세시대', '법률저널', '베네핏 매거진', '베리타스알파',
              '베스트베이비', '베스트일레븐', '베이커리', '벤처스퀘어', '벤치월드', '벨리시마', '보드나라', '보안24', '보험신보', '복지TV', '복지닷컴', '봉제기술',
              '부천포커스', '북데일리', '불교저널', '뷰티경제', '뷰티누리', '뷰티라이프', '뷰티엠', '뷰티투데이', '뷰티패션', '뷰티한국', '브레인박스', '브레인스톰',
              '브릿지경제', '브이알앤', '블랙벨트코리아', '블로터닷넷', '비마이너', '비전시스템', '비즈니스리포트', '비즈니스워치', '비즈니스포스트', '비즈니스플러스',
              '비즈볼프로젝트', '비즈엔터', '비즈트리뷴', '비즈한국', '비트허브', '사건의 내막', '사건의내막', '사람과산', '사이언스엠디', '사이언스올', '사이클TV', '사진예술',
              '사커뱅크', '사회복지동행', '산업데일리', '상용차매거진', '생활성서', '서예문화', '서울21', '서울경제 골프매거진트', '서울경제TV', '서울대동창회보', '서울매일',
              '서울문화투데이', '서울아트가이드', '서울파이낸스', '서울포스트', '설비기술', '세계거리축제', '세계닷컴', '세계파이낸스', '세라믹 코리아', '세미나투데이',
              '세이프투데이', '소년중앙', '소비라이프', '소비자TV', '소비자연합타임스', '솔라투데이', '수상레저', '수학동아', '숙박매거진', '쉬즈라이프', '쉽빌딩',
              '스마트PC사랑', '스카우트', '스카이데일리', '스쿨iTV', '스크랩워치', '스타에이지', '스타일 에이치', '스타저널', '스타투데이', '스타패션', '스타포커스',
              '스틸데일리', '스틸마켓', '스틸프라이스', '스페셜경제', '스포츠Q', '스포츠경향', '스포츠동아', '스포츠서울', '스포츠온', '스포츠월드', '스포츠조선', '스포츠코리아',
              '스포츠타임스', '스포츠투데이', '스포츠한국', '스포탈코리아', '스포티안', '시니어조선', '시사IN', '시사골프', '시사메디in', '시사오늘', '시사위크', '시사저널',
              '시사저널e', '시사주간', '시사코리아', '시사코리아저널', '시사타임', '시사포커스', '시티라이프', '시흥저널', '식품외식경제', '식품저널', '신동아',
              '신재생에너지저널', '신제품신기술', '실버TV', '쎄씨캠퍼스', '씨네21', '씨넷코리아', '아레나 옴므 플러스', '아레나옴므플러스', '아름다운동행', '아리랑TV',
              '아시아경제', '아시아경제TV', '아시아글로브', '아시아에너지경제', '아시아투데이', '아웃도어', '아웃소싱타임스', '아웃스탠딩', '아이가스저널', '아이데일리',
              '아이티데일리', '아이티비즈', '아주경제', '아츠앤컬쳐', '아쿠아인포', '아트앤디자인', '아트앤씨', '아트앤컬렉터', '아트인컬처', '아트저널', '아트코리아방송',
              '아트프라이스', '안전리뷰', '안전저널', '안전정보', '앙쥬', '애니멀매거진', '애슬릿미디어', '애플경제', '약사공론', '어린이과학동아', '어린이동아', '얼루어 코리아',
              '얼리어답터', '업코리아', '에너지데일리', '에너지설비관리', '에너지코리아', '에듀데일리', '에듀동아', '에듀진', '에브리골프', '에스코트서울', '에스콰이어',
              '에이비 로드', '에코데일리', '에코스케이프', '에코저널', '에코타임스', '엔사이드', '엔지니어링데일리', '엔카매거진', '엔터미디어', '엘리베이터·주차설비', '여성동아',
              '여성조선', '여성중앙', '여수투데이', '여행스케치', '연합인포맥스', '열관리시공', '예술TV Arte', '예술세계', '예천저널', '오디오매거진', '오뜨피플',
              '오토다이어리', '오토데일리', '오토레이싱', '오토모닝', '오토모빌썬', '오토모티브 일렉트로닉스', '오토뷰', '오토사운드', '오토카 코리아', '오토캠핑', '오토헤럴드',
              '오픈모바일', '온케이웨더', '올포칩 미디어', '와이드스포츠', '와이드커버리지', '와인리뷰', '요트피아', '용접저널', '우드플래닛', '우등생과학', '우등생논술',
              '우등생학습', '우먼데일리', '우먼라이프', '우먼센스', '우먼컨슈머', '울진마당', '워터저널', '월간 HRD', '월간 가구가이드', '월간 객석', '월간 건축사',
              '월간 낚시21', '월간 도예', '월간 디자인', '월간 문화재', '월간 미대입시', '월간 미술', '월간 붕어', '월간 비디오플러스', '월간 사진', '월간 산',
              '월간 서예', '월간 수퍼마켓', '월간 안과정보', '월간 암', '월간 앱', '월간 외식경영', '월간 윈도어', '월간 전시가이드', '월간 주얼리앤워치', '월간 주유소',
              '월간 중앙', '월간 중장비', '월간 창호기술', '월간 탁구', '월간 포장', '월간 포토닷', '월간 항공', '월간 헌정', '월간 홈쇼핑', '월간CEO', '월간교육',
              '월간에세이', '월간유아', '월간인물', '월간인테리어', '월간전기', '월간조선', '월간커피', '월드얀', '웨딩21', '웰니스투데이', '웰스매니지먼트', '웹브릿지',
              '위기관리경영', '위드레저', '위즈키즈', '위클리서울', '위클리오늘', '위키트리', '유니타스브랜드', '유스라인', '유체제어', '유통데일리', '음식과사람', '음악저널',
              '음악춘추', '의료기기협회보', '의료정보', '이데일리', '이데일리TV', '이버즈', '이슈데일리', '이지경제', '이코노미 인사이트', '이코노미21', '이코노미TV',
              '이코노미스트', '이코노미조선', '이코노믹리뷰', '이코노믹포스트', '이투데이', '익스테리어', '인권하루소식', '인베스트조선', '인사관리', '인사이드저널', '인사이드코리아',
              '인성시대', '인쇄문화', '인스정보기술', '인스타일', '인재경영', '인터넷토마토', '인터풋볼', '일간보사', '일간부동산', '일간스포츠', '일간조선해양', '일간투데이',
              '일간환경', '일렉트릭파워', '일요경제', '일요서울', '일요시사', '자동인식·보안', '자동제어계측', '자동차생활', '자동차와주유소', '자동차제조기술', '자전거생활',
              '작은것이아름답다', '장애인문화방송', '전기기술', '전기설비', '전기저널', '전기평론', '전력경제', '전북포스트', '전원생활', '전원속의 내집', '전원주택라이프',
              '전자과학', '전자기술', '전자부품', '전주교차로', '점프볼', '정읍투데이', '제주저널', '조명과인테리어', '조선닷컴', '조선닷컴 더스타', '조선비즈', '조선에듀',
              '조인스닷컴', '조인스랜드부동산', '좋은교육미디어', '주간 프랜차이즈', '주간경향', '주간동아', '주간무역', '주간조선', '주간코스메틱', '주간한국', '주간현대',
              '주단조와열처리', '주류저널', '주택저널', '중기이코노미', '중부매일', '중앙선데이', '증권플러스', '지디넷 코리아', '참여와혁신', '참좋은환경', '창과문',
              '창업&프랜차이즈', '채널A', '채널프랜차이즈', '철구기술', '첨단환경기술', '청년의사', '초이스경제', '축구저널', '출판문화', '출판저널', '춤과사람들', '춤판춤북',
              '충남투데이', '충청매일', '충청투데이', '치프 이그제큐티브', '카고프레스', '카리포트', '카매거진', '카모드', '카미디어', '카스카미디어', '카앤모델', '카앤스포츠',
              '카오디오', '카조선', '카테크', '카포스', '카포탈넷', '카홀릭', '캐드앤그래픽스', '캠퍼스 잡앤조이', '컨슈머리서치', '컨슈머와이드', '컨슈머치', '컨슈머타임스',
              '컬처오션', '컬처윈도우', '케이벤치', '켐로커스', '코리아닷컴', '코리아데일리', '코리아본드웹', '코리아중앙데일리', '코리아타임스', '코리아트래블', '코리아헤럴드',
              '코메디닷컴', '코스모캠퍼스', '코스모폴리탄', '코어플라넷', '쿠켄', '크레이닷티비', '크리스천투데이', '탑골프', '탑라이더', '테니스코리아', '테니스피플', '테크엠',
              '테크홀릭', '텐아시아', '토마토TV', '토요경제', '토이매거진', '투데이안', '투데이에너지', '투데이코리아', '투어코리아', '트래블데일리', '트래블레저플러스',
              '트래블맵', '트래블아이', '트래블조선', '트랙사이드', '트레블 라이프', '트루스토리', '티매거진', '티몬매거진', '티브이데일리', '파운드매거진', '파워코리아',
              '파워코리아데일리', '파이낸셜투데이', '파이낸스투데이', '파퓰러사이언스', '팍스넷데일리', '팜스탁', '패션리뷰', '패션비즈', '패션서울', '패션웹진 스냅', '패션인사이트',
              '패션저널', '패션채널', '팩트TV', '퍼포먼스텐', '펌프기술', '펫저널', '포모스', '포브스코리아', '포장기계', '포춘코리아', '포커스경제', '포포투', '폴리진',
              '표면실장기술', '표면처리저널', '푸드TV', '푸드트래블', '품질경영', '풋볼리스트', '퓨쳐에코', '프라임경제', '프런티어타임스', '프레시안', '프레지던트', '프로슈머',
              '프리빌리지', '프린팅코리아', '플라스틱사이언스', '플라스틱월드', '플라스틱코리아', '플래텀', '플랜트기술', '하우징헤럴드', '하이닥', '하이맘', '하퍼스 바자',
              '한겨레21', '한경닷컴', '한경닷컴 게임톡', '한경닷컴 스타엔', '한경비즈니스', '한국경제', '한국경제TV', '한국수산경제', '한국스포츠경제', '한국아이닷컴',
              '한국여행업협회', '한국연극', '한국의약통신', '한국종합기술', '한옥문화', '함께걸음', '함께사는길', '항공문화', '해양과조선', '해양한국', '해저여행', '해피CGI',
              '핸드볼코리아', '행복이가득한집', '허핑턴포스트코리아', '헝그리앱', '헤드라이너', '헤럴드POP', '헤럴드경제', '헤렌', '헬스경향', '헬스앤라이프', '헬스조선',
              '현대경영', '현대해양', '호텔&레스토랑', '화이트페이퍼', '화학장치기술', '환경과조경', '환경기술인', '환경데일리', '환경미디어', '환경방송', '환경정보',
              '환경포커스', '효도실버', '후생신보', '히트펌프·공조', '모바일 경향']

media_suffix = [
    "신문", "뉴스", "일보", "타임즈", "타임스"
]

split_words = [
    f"{prefix}{word}" for word in
    ["▶", "▷", "▶", "▷", "▸", "▹", "►", "▻", "◂", "◃", "◄", "◅",
     "■", "□", "▣", "▤", "▥", "▦", "▧", "▨", "▩", "▪", "▫", "☐",
     "◆", "◇", "◈", "◉", "◊", "◐", "◑", "◒", "◓",
     "©", "ⓒ", "☛", "☞", "☚", "☜", "☎", "☏",
     "Copyright", "저작권자", "무단전재", "무단 전재", "(끝)",
     "기사제보", "보도자료", "제보하기", "기사문의", "많이 본 기사", "취재본부",
     "촬영기자", "영상편집", "카카오톡:", "카카오톡 :", "전화:", "전화 :"] + media_list
    for prefix in ["<", "[", "(", ""]
]

split_patterns = [
    re.compile(
        r"(?i)(\b(지금까지|.{2,7}에서)\b)\W+(?:\w+[^\w]+){0,6}?\b(기자|특파원|교통정보|KBS 뉴스|SBS 뉴스|MBC 뉴스|YTN|MBN|뉴스더하기)"
    ),
    re.compile(
        r"(?i)(\b(KBS 뉴스|SBS 뉴스|MBC 뉴스|YTN|MBN|뉴스더하기)\b)\W+(?:\w+){0,3}?(였습니다\.|입니다\.).*"
    ),
    re.compile("|".join([re.escape(p) for p in split_words])),
    re.compile("기자 | 기자| 특파원|특파원 |@"),
]

remove_patterns = {
    re.compile(r'\([^)]*(=| 기자| 특파원|뉴스|뉴스1|신문|일보|타임즈|타임스)[^)]*\)'): "",
    re.compile(r"\s*[\w가-힣·=]*\s*(기자=|기자 =|특파원=|특파원 =)"): "",
    re.compile(r"\s*[\w가-힣·=]*\s*기자|특파원"): "",
    re.compile("(기자 =|특파원 =|기자=|특파원=)|AP=연합뉴스|=연합뉴스|"): "",
    re.compile("=뉴스1"): "",
    re.compile(r"【.*?】|\[.*?]"): "",
    re.compile(r"[a-zA-Z0-9]+@"): "",
}

email_pattern = re.compile(r"[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}")
do_not_split_pattern = re.compile("기자[가-힣]")


def _split_spam(sents, found_media):
    sents_for_use = []
    for sent in sents:
        drop_sent = False
        for check_word, found_word in found_media.items():
            if sent.endswith(found_word):
                drop_sent = True
                break
            elif f"{check_word} " in sent or f"{check_word}\n" in sent:
                drop_sent = True
                break

        if not drop_sent:
            for pattern in split_patterns:
                if pattern.search(sent):
                    do_not_split = False
                    if do_not_split_pattern.search(sent):
                        do_not_split = True

                    if not do_not_split:
                        drop_sent = True
                        break

        if drop_sent:
            continue
        else:
            sents_for_use.append(sent)

    return sents_for_use


def _replace_spam(sents):
    sents_for_use = []
    for sent in sents:
        do_not_split = False
        if do_not_split_pattern.search(sent):
            do_not_split = True

        if not do_not_split:
            for pattern, repl in remove_patterns.items():
                sent = pattern.sub(repl, sent)
        sents_for_use.append(sent)
    return sents_for_use


def _should_skip(text):
    for bad in skip_words:
        if bad in text:
            return True
    return False


def _normalize_text(text, postprocess):
    if not postprocess:
        for before, after in normalization_open.items():
            while before in text:
                text = text.replace(before, after)
        for before, after in normalization_close.items():
            while before in text:
                text = text.replace(before, after)

    for before, after in normalization_others.items():
        while before in text:
            text = text.replace(before, after)
    return text


def _find_media(text):
    words = text.split()
    media_found = {}

    for word in words:
        for media in media_suffix:
            if media in word:
                if media not in media_found:
                    media_found[media] = [word]
                else:
                    media_found[media].append(word)

    return {
        media: f"{prefix}{word}"
        for media, word in media_found.items()
        for prefix in ["<", "[", "(", "< ", "[ ", "( ", ""]
    }


def _split_sentences(text):
    sents = []
    for sent in re.split(r"(?<=[.!?]\s)", text):
        for s in re.split(r"(?<=[다요죠][.!?])(?![\"'\])>])", sent):
            if len(s.strip()) > 0:
                sents.append(s)
    return sents


def _pre_split(text):
    text = text.split("☞")[0]
    text = email_pattern.split(text)[0]
    return text


def clean_news(
    text: Union[str, List[str], Tuple[str]],
    min_sentences: int = 3,
    header_ratio: float = 0.4,
    footer_ratio: float = 0.4,
    num_workers: Union[int, str] = "auto",
    verbose: bool = False,
) -> Union[str, List[str]]:
    """
    This cleans news articles by removing useless headers and footers.

    Args:
        text (Union[str, List[str], Tuple[str]]): Input text or list of texts.
        min_sentences (int): Minimum number of sentences to keep. Defaults to 3.
        header_ratio (float): Ratio of the number of sentences to check in the header. Defaults to 0.4.
        footer_ratio (float): Ratio of the number of sentences to check in the footer. Defaults to 0.4.
        num_workers (Union[int, str]): the number of multiprocessing workers
        verbose (bool): whether to print verbose outputs or not

    Returns:
        Union[str, List[str]]: Cleaned text or list of cleaned texts.

    Examples:
        >>> from kss import Kss
        >>> clean_news = Kss("clean_news")
        >>> text = "[사진]에버랜드, 봄꽃 펼쳐진 '튤립축제' 오픈\\n\\n[ 뉴스1 제공](서울=뉴스1) 이동원 기자 = 에버랜드가 오는 22일부터 봄을 상징하는 튤립 120만 송이와 함께 '튤립축제'를 오픈해 본격적인 봄의 시작을 알린다. 지난 1992년 국내 첫 튤립 축제를 연 이후 올해로 22회째를 맞이한 에버랜드 '튤립축제'는 지난해 첫 선을 보이며 좋은 반응을 얻었던 오감(五感)체험 '시크릿가든'을 리뉴얼하고, 신규 테마 꽃길을 조성하는 등 봄꽃을 활용한 다양한 볼거리를 강화한 것이 특징이다. 또한 4월 28일까지 열리는 '튤립축제'에서는 야간 개장과 함께 손님 참여요소가 늘어난 인기 공연, 퍼레이드가 재오픈하는 등 봄을 맞아 나들이 나온 상춘객들의 눈과 귀를 즐겁게 할 예정이다. (에버랜드 제공)2013.3.10/뉴스1 < 저작권자 뉴스1 코리아, 무단전재 및 재배포 금지 > ☞ 뉴스1 바로가기 [스타뉴스 공식 글로벌 버전 애플리케이션] [증권알리미]국내외 증시핫이슈 및 오늘의 승부주! [머니원]北 리스크로 조정, 매수 기회로 [머니투데이 핫뉴스] ☞ 신입사원 재테크, 5년 안에 '1억 만들기' ☞ '박시후·박준' 고소女, 둘다 '여기'가더니 ☞ 늙으면 '돈이라도' 있어야 하는 진짜 이유 ☞ 손연재 가방 시끌…'신입생은 명품 안돼?' ☞ '바람난 부인 뒷조사' 300만원 줬더니… [book]10년의 선택, 중국에 투자하라 [핫이슈]'멘사' 천재들 뭉쳐 15년 투자했는데... '맙소사' 뉴스1 제공 < 저작권자 ⓒ '돈이 보이는 리얼타임 뉴스' 머니투데이, 무단전재 및 재배포 금지"
        >>> output = clean_news(text)
        >>> print(output)
        '에버랜드, 봄꽃 펼쳐진 '튤립축제' 오픈\\n\\n에버랜드가 오는 22일부터 봄을 상징하는 튤립 120만 송이와 함께 '튤립축제'를 오픈해 본격적인 봄의 시작을 알린다. 지난 1992년 국내 첫 튤립 축제를 연 이후 올해로 22회째를 맞이한 에버랜드 '튤립축제'는 지난해 첫 선을 보이며 좋은 반응을 얻었던 오감(五感)체험 '시크릿가든'을 리뉴얼하고, 신규 테마 꽃길을 조성하는 등 봄꽃을 활용한 다양한 볼거리를 강화한 것이 특징이다. 또한 4월 28일까지 열리는 '튤립축제'에서는 야간 개장과 함께 손님 참여요소가 늘어난 인기 공연, 퍼레이드가 재오픈하는 등 봄을 맞아 나들이 나온 상춘객들의 눈과 귀를 즐겁게 할 예정이다.'
    """
    text, finish = _check_text(text)

    if finish:
        return text

    min_sentences = _check_type(min_sentences, "min_sentences", int)
    header_ratio = _check_type(header_ratio, "header_ratio", float)
    footer_ratio = _check_type(footer_ratio, "footer_ratio", float)
    assert 0 <= header_ratio <= 1, "header_ratio should be in [0, 1]"
    assert 0 <= footer_ratio <= 1, "footer_ratio should be in [0, 1]"
    assert header_ratio + footer_ratio < 1, "header_ratio + footer_ratio should be less than 1"
    verbose = _check_type(verbose, "verbose", bool)
    num_workers = _check_num_workers(text, num_workers)

    if num_workers is not False and verbose:
        verbose = False
        logger.warn(
            "Verbose mode is not supported for multiprocessing. "
            "It will be turned off automatically."
        )

    return _run_job(
        func=partial(
            _clean_news,
            min_sentences=min_sentences,
            header_ratio=header_ratio,
            footer_ratio=footer_ratio,
            verbose=verbose,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _clean_news(
    text: str,
    min_sentences: int = 3,
    header_ratio: float = 0.4,
    footer_ratio: float = 0.4,
    verbose: bool = False,
):
    input_text = text
    skip_sample = False
    # for debug.

    if _should_skip(text):
        skip_sample = True

    if not skip_sample:
        text = _normalize_text(text, postprocess=False)
        text = _pre_split(text)
        sents = _split_sentences(text)

        if len(sents) <= min_sentences:
            skip_sample = True
        else:
            found_media = _find_media(text)

            if len(sents) <= min_sentences * 2:
                sents = _replace_spam(sents)
                sents = _split_spam(sents, found_media)

                if len([s for s in sents if len(s.strip()) > 0]) == 0:
                    skip_sample = True
                else:
                    text = " ".join(sents)
                    text = _normalize_text(text, postprocess=True).strip()

            else:
                num_header_sent = int(len(sents) * header_ratio)
                num_footer_sent = int(len(sents) * footer_ratio)

                header_sents = sents[:num_header_sent]
                middle_sents = sents[num_header_sent:-num_footer_sent]
                footer_sents = sents[-num_footer_sent:]

                header_sents = _replace_spam(header_sents)
                footer_sents = _split_spam(footer_sents, found_media)
                text = " ".join(header_sents + middle_sents + footer_sents)
                text = _normalize_text(text, postprocess=True).strip()

    if len(text.strip()) == 0:
        skip_sample = True

    if skip_sample:
        output_text = None
    else:
        output_text = text

    if verbose:
        print(highlight_diffs(input_text, output_text).replace("\n", "\\n"))

    return output_text



import multiprocessing as mp
from typing import Callable, Any, Optional, Union, List, Tuple

import tossi

from kss._modules.morphemes.split_morphemes import split_morphemes


def _check_text(
    param_1: Union[str, List[str]],
    param_2: Union[str, List[str]],
    param_1_name: str,
    param_2_name: str
) -> Tuple[Union[str, List[str]], Union[str, List[str]]]:
    if isinstance(param_1, str) and isinstance(param_2, str):
        if len(param_1) == 0:
            raise ValueError(f"`{param_1_name}` must not be empty")
        if len(param_2) == 0:
            raise ValueError(f"`{param_2_name}` must not be empty")
        return [param_1], [param_2]

    elif isinstance(param_1, (list, tuple)) and isinstance(param_2, (list, tuple)):
        if len(param_1) != len(param_2):
            raise ValueError(
                f"`{param_1_name}` and `{param_2_name}` must have the same length when both are lists or tuples")
        if any(len(p) == 0 for p in param_1):
            raise ValueError(f"`{param_1_name}` must not contain an empty string")
        if any(len(j) == 0 for j in param_2):
            raise ValueError(f"`{param_2_name}` must not contain an empty string")
        return param_1, param_2

    elif isinstance(param_1, (list, tuple)) and isinstance(param_2, str):
        if any(len(p) == 0 for p in param_1):
            raise ValueError(f"`{param_1_name}` must not contain an empty string")
        if len(param_2) == 0:
            raise ValueError(f"`{param_2_name}` must not be empty")
        return param_1, [param_2] * len(param_1)

    elif isinstance(param_1, str) and isinstance(param_2, (list, tuple)):
        if len(param_1) == 0:
            raise ValueError(f"`{param_1_name}` must not be empty")
        if any(len(j) == 0 for j in param_2):
            raise ValueError(f"`{param_2_name}` must not contain an empty string")
        return [param_1] * len(param_2), param_2

    else:
        raise ValueError(f"`{param_1_name}` and `{param_2_name}` must be strings or lists/tuples of strings")


def _run_job(
    func: Callable,
    input_1: List[str],
    input_2: List[str],
    input_1_name: str,
    input_2_name: str,
    num_workers: Optional[Union[int, bool]] = None,
) -> Union[Any, List[Any]]:
    """
    Run job with or without multiprocessing.

    Args:
        func (Callable): function to run
        input_1 (List[str]): input data
        input_2 (List[str]): input data
        input_1_name (str): input data name
        input_2_name (str): input data name
        num_workers (Optional[Union[int, bool]]): the number of multiprocessing workers.
    """
    if num_workers is False:
        if len(input_1) == 1 and len(input_2) == 1:
            return func(input_1[0], input_2[0])
        elif len(input_1) == 1 and len(input_2) > 1:
            return [func(input_1[0], j) for j in input_2]
        elif len(input_1) > 1 and len(input_2) == 1:
            return [func(i, input_2[0]) for i in input_1]
        else:
            return [func(i, j) for i, j in zip(input_1, input_2)]
    else:
        with mp.Pool(num_workers) as pool:
            if len(input_1) == 1 and len(input_2) == 1:
                raise ValueError(
                    f"Oops! `{input_1_name}` or `{input_2_name}` "
                    f"must have at least 2 elements when using multiprocessing."
                )
            elif len(input_1) == 1 and len(input_2) > 1:
                output = pool.starmap(func, [(input_1[0], j) for j in input_2])
            elif len(input_1) > 1 and len(input_2) == 1:
                output = pool.starmap(func, [(i, input_2[0]) for i in input_1])
            else:
                output = pool.starmap(func, list(zip(input_1, input_2)))

            pool.close()
            pool.join()
            return output


def _check_num_workers(
    param_1: List[str],
    param_2: List[str],
    param_1_name: str,
    param_2_name: str,
    num_workers: Union[int, str],
) -> Optional[Union[int, bool]]:
    if isinstance(num_workers, float):
        num_workers = int(num_workers)
    elif isinstance(num_workers, str):
        num_workers = num_workers.lower()

    if not isinstance(num_workers, int) and num_workers != "auto":
        raise TypeError(
            f"Oops! '{num_workers}' is not supported value for `num_workers`.\n"
            f"Currently kss only supports [int, 'auto'] for this.\n"
            f"Please check `num_workers` parameter again ;)"
        )
    elif isinstance(num_workers, int) and num_workers < 1:
        raise ValueError(
            f"Oops! `num_workers` must be same or greater than 1, but you input {num_workers}.\n"
            "- You can input 1 for no multiprocessing.\n"
            "- You can input 2 ~ n for multiprocessing.\n"
            "- You can input 'auto' for using maximum workers.\n"
            "Please check `num_workers` parameter again ;)"
        )

    if len(param_1) == 1 and len(param_2) == 1:
        return False
        # no multiprocessing worker

    elif num_workers == "auto":
        num_workers = None
        # maximum multiprocessing workers

    return num_workers


def _preprocess_josa(prefix, josa):
    if "도록" in josa or "토록" in josa:
        # 공부하다와 같은 동사(v)의 경우, '~도록'을 사용합니다.
        # 평생, 영원과 같은 명사(n)의 경우, '~토록'을 사용합니다.
        # https://www.italki.com/ko/post/question-468087
        return "토록" if split_morphemes(prefix)[-1][1].startswith("N") else "도록"

    elif "던" in josa and (prefix.endswith("어쨋") or prefix.endswith("어쨌") or prefix.endswith("어쩌")):
        # '어쨌든'은 앞의 상황과는 관계 없이란 뜻으로 '어찌하였든'이 줄어진 말입니다.
        # 이때 '-든'을 '-던'으로 쓰는 것은 잘못입니다. '-던'의 '-더-'는 회상을 나타내는 말이므로 과거의 일을 이야기할 때 가능합니다.
        # 그런데 '어쨌든'은 과거의 일을 이야기하지 않는 것이므로 '-던'으로 쓸 이유가 없는 것입니다.
        # https://m.cafe.daum.net/bareunNGO/5Dfn/55
        return "든"

    elif "에게" in josa and (prefix.endswith("것") or prefix.endswith("곳") or prefix.endswith("거")):
        # ‘에’와 ‘에게’를 쓸 때는 그 조사 앞에 오는 말이 ‘유정 명사’냐 아니냐에 따라 달라집니다. ‘유정 명사’라는 것은 사람이나 동물을 가리키는 명사를 말합니다.
        # 모든 유정/무정 명사를 가릴 수 없으니 대표적인 '것', '곳', '거'만 처리.
        # https://world.kbs.co.kr/service/contents_view.htm?lang=k&menu_cate=learnkorean&id=&board_seq=228909&page=249
        return josa.replace("에게", "에")

    return josa


def _select_josa(prefix, josa):
    prefix = prefix.strip()
    josa = josa.strip()

    josa = _preprocess_josa(prefix, josa)
    return tossi.pick(prefix, josa)


def _combine_josa(prefix, josa):
    return prefix + _select_josa(prefix, josa)



# This code is copied from korean-romanizer [https://github.com/osori/korean-romanizer]
# And modified by Hyunwoong Ko [https://github.com/hyunwoongko] and Sohyeon Yim [https://github.com/sohyunwriter]

import re
from functools import partial
from typing import Union, List, Tuple

from unidecode import unidecode

from kss._modules.g2p.g2p import g2p
from kss._modules.romanization.utils import pronounce, Syllable
from kss._utils.multiprocessing import _run_job
from kss._utils.sanity_checks import _check_text, _check_type, _check_num_workers, _check_backend_mecab_pecab_only

vowel = {
    # 단모음 monophthongs
    'ㅏ': 'a',
    'ㅓ': 'eo',
    'ㅗ': 'o',
    'ㅜ': 'u',
    'ㅡ': 'eu',
    'ㅣ': 'i',
    'ㅐ': 'ae',
    'ㅔ': 'e',
    'ㅚ': 'oe',
    'ㅟ': 'wi',
    # 이중모음 diphthongs
    'ㅑ': 'ya',
    'ㅕ': 'yeo',
    'ㅛ': 'yo',
    'ㅠ': 'yu',
    'ㅒ': 'yae',
    'ㅖ': 'ye',
    'ㅘ': 'wa',
    'ㅙ': 'wae',
    'ㅝ': 'wo',
    'ㅞ': 'we',
    'ㅢ': 'ui',  # [붙임 1] ‘ㅢ’는 ‘ㅣ’로 소리 나더라도 ‘ui’로 적는다.
}

# 초성 Choseong (Syllable Onset)
onset = {
    # 파열음 stops/plosives
    'ᄀ': 'g',
    'ᄁ': 'kk',
    'ᄏ': 'k',
    'ᄃ': 'd',
    'ᄄ': 'tt',
    'ᄐ': 't',
    'ᄇ': 'b',
    'ᄈ': 'pp',
    'ᄑ': 'p',
    # 파찰음 affricates
    'ᄌ': 'j',
    'ᄍ': 'jj',
    'ᄎ': 'ch',
    # 마찰음 fricatives
    'ᄉ': 's',
    'ᄊ': 'ss',
    'ᄒ': 'h',
    # 비음 nasals
    'ᄂ': 'n',
    'ᄆ': 'm',
    # 유음 liquids
    'ᄅ': 'r',
    # Null sound
    'ᄋ': '',
}

coda = {
    # 파열음 stops/plosives
    'ᆨ': 'k',
    'ᆮ': 't',
    'ᆸ': 'p',
    # 비음 nasals
    'ᆫ': 'n',
    'ᆼ': 'ng',
    'ᆷ': 'm',
    # 유음 liquids
    'ᆯ': 'l',
    None: '',
}


def romanize(
    text: Union[str, List[str], Tuple[str]],
    use_morpheme_info: bool = True,
    backend: str = "auto",
    convert_english_to_hangul_phonemes: bool = False,
    convert_numbers_to_hangul_phonemes: bool = False,
    num_workers: Union[int, str] = "auto",
) -> Union[str, List[str]]:
    """
    This romanizes Korean text.

    Args:
        text (Union[str, List[str], Tuple[str]]): single text or list of texts
        use_morpheme_info (bool): whether to use morpheme information or not
        backend (str): morpheme analyzer backend. 'mecab', 'pecab' are supported
        convert_english_to_hangul_phonemes (bool): whether to convert English to Hangul phonemes or not
        convert_numbers_to_hangul_phonemes (bool): whether to convert numbers to Hangul phonemes or not
        num_workers (Union[int, str]): the number of multiprocessing workers

    Returns:
        Union[str, List[str]]: romanized text or list of romanized texts

    Examples:
        >>> from kss import Kss
        >>> romanize = Kss("romanize")
        >>> text = "안녕하세요"
        >>> romanize(text)
        'annyeonghaseyo'
        >>> text = "대관령"
        >>> romanize(text)
        'daegwallyeong'

    References:
        This was copied from [korean-romanizer](https://github.com/osori/korean-romanizer) and modified by Kss
    """

    text, finish = _check_text(text)

    if finish:
        return text

    use_morpheme_info = _check_type(use_morpheme_info, "use_morpheme_info", bool)
    _check_backend_mecab_pecab_only(backend)
    convert_english_to_hangul_phonemes = _check_type(convert_english_to_hangul_phonemes,
                                                     "convert_english_to_hangul_phonemes", bool)
    convert_numbers_to_hangul_phonemes = _check_type(convert_numbers_to_hangul_phonemes,
                                                     "convert_numbers_to_hangul_phonemes", bool)
    num_workers = _check_num_workers(text, num_workers)

    return _run_job(
        func=partial(
            _romanize,
            use_morpheme_info=use_morpheme_info,
            backend=backend,
            convert_english_to_hangul_phonemes=convert_english_to_hangul_phonemes,
            convert_numbers_to_hangul_phonemes=convert_numbers_to_hangul_phonemes,
        ),
        inputs=text,
        num_workers=num_workers,
    )


def _romanize(
    text,
    use_morpheme_info=True,
    backend="auto",
    convert_english_to_hangul_phonemes=False,
    convert_numbers_to_hangul_phonemes=False,
):
    if use_morpheme_info:
        pronounced = g2p(
            text,
            backend=backend,
            convert_english_to_hangul_phonemes=convert_english_to_hangul_phonemes,
            convert_numbers_to_hangul_phonemes=convert_numbers_to_hangul_phonemes,
        )
    else:
        pronounced = pronounce(
            text,
            convert_english_to_hangul_phonemes=convert_english_to_hangul_phonemes,
            convert_numbers_to_hangul_phonemes=convert_numbers_to_hangul_phonemes,
        )

    romanized_text = ""
    for char in pronounced:
        if re.match(r"[가-힣ㄱ-ㅣ]", char):
            s = Syllable(char)

            if not s.medial and not s.final:
                # s is NOT a full syllable (e.g. characters)
                if onset.get(chr(s.initial)):
                    romanized_text += onset[chr(s.initial)]
                elif vowel.get(chr(s.initial)):
                    romanized_text += vowel[chr(s.initial)]
                else:
                    romanized_text += unidecode(char)
            else:
                # s is a full syllable
                romanized_text += onset[s.initial] + vowel[s.medial] + coda[s.final]

        else:
            romanized_text += char

    # 유음화 / 비음화 처리
    if use_morpheme_info:
        # 울릉도 ulreungdo -> ulleungdo
        # 대관령 daegwalryeong -> daegwallyeong
        romanized_text = romanized_text.replace("lr", "ll")

    return romanized_text



from kss import Kss


def test_split_hanja():
    split_hanja = Kss("split_hanja")
    text = "大韓民國은 民主共和國이다."
    output = split_hanja(text)
    assert output == ["大韓民國", "은 ", "民主共和國", "이다."]


def test_is_hanja():
    is_hanja = Kss("is_hanja")
    text = "大"
    output = is_hanja(text)
    assert output is True
    text = "대"
    output = is_hanja(text)
    assert output is False


def test_hanja2hangul():
    hanja2hangul = Kss("hanja2hangul")
    text = "大韓民國은 民主共和國이다."
    output = hanja2hangul(text)
    assert output == "대한민국은 민주공화국이다."


if __name__ == '__main__':
    test_split_hanja()
    test_is_hanja()
    test_hanja2hangul()
